# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
A plugin for drawing canvas forms (overlaid graphics).

**Plugin Type: Local**

``Drawing`` is a local plugin, which means it is associated with a
channel.  An instance can be opened for each channel.

**Usage**

This plugin can be used to draw many different shapes on the image display.
When it is in "draw" mode, select a shape from the drop-down menu, adjust
the shape's parameters (if needed), and draw on the image by using left
mouse button. You can choose to draw in pixel or WCS space.

To move or edit an existing shape, set the plugin on "edit" or "move" mode,
respectively.

To save the drawn shape(s) as mask image, click the "Create Mask" button
and you will see a new mask image created in Ginga. Then, use ``SaveImage``
plugin to save it out as single-extension FITS. Note that the mask will
take the size of the displayed image. Therefore, to create masks for
different image dimensions, you need to repeat the steps multiple times.

"""
from datetime import datetime

from ginga import GingaPlugin
from ginga import colors
from ginga.gw import Widgets
from ginga.misc import ParamSet, Bunch
from ginga.util import dp
from ginga.canvas.CanvasObject import coord_names

__all__ = ['Drawing']

draw_colors = colors.get_colors()

default_drawtype = 'circle'
default_drawcolor = 'lightblue'
fillkinds = ('circle', 'rectangle', 'polygon', 'triangle', 'righttriangle',
             'square', 'ellipse', 'box')


class Drawing(GingaPlugin.LocalPlugin):

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super(Drawing, self).__init__(fv, fitsimage)

        self.layertag = 'drawing-canvas'

        self.dc = fv.get_draw_classes()
        canvas = self.dc.DrawingCanvas()
        canvas.enable_draw(True)
        canvas.enable_edit(True)
        canvas.set_drawtype('point', color='cyan')
        canvas.set_callback('draw-event', self.draw_cb)
        canvas.set_callback('edit-event', self.edit_cb)
        canvas.set_callback('edit-select', self.edit_select_cb)
        canvas.set_surface(self.fitsimage)
        # So we can draw and edit with the cursor
        canvas.register_for_cursor_drawing(self.fitsimage)
        self.canvas = canvas

        self.drawtypes = list(canvas.get_drawtypes())
        self.drawcolors = draw_colors
        self.linestyles = ['solid', 'dash']
        self.coordtypes = coord_names
        # contains all parameters to be passed to the constructor
        self.draw_args = []
        self.draw_kwdargs = {}
        # cache of all canvas item parameters
        self.drawparams_cache = {}
        # holds object being edited
        self.edit_obj = None

        # For mask creation from drawn objects
        self._drawn_tags = []
        self._mask_prefix = 'drawing'

    def build_gui(self, container):
        top = Widgets.VBox()
        top.set_border_width(4)

        vbox, sw, orientation = Widgets.get_oriented_box(container)
        self.orientation = orientation
        vbox.set_border_width(4)
        vbox.set_spacing(2)

        fr = Widgets.Frame("Drawing")

        captions = (("Draw type:", 'label', "Draw type", 'combobox'),
                    )
        w, b = Widgets.build_info(captions)
        self.w.update(b)

        combobox = b.draw_type
        for name in self.drawtypes:
            combobox.append_text(name)
        index = self.drawtypes.index(default_drawtype)
        combobox.set_index(index)
        combobox.add_callback(
            'activated', lambda w, idx: self.set_drawtype_cb())

        fr.set_widget(w)
        vbox.add_widget(fr, stretch=0)

        mode = self.canvas.get_draw_mode()
        hbox = Widgets.HBox()
        btn1 = Widgets.RadioButton("Draw")
        btn1.set_state(mode == 'draw')
        btn1.add_callback(
            'activated', lambda w, val: self.set_mode_cb('draw', val))
        btn1.set_tooltip("Choose this to draw")
        self.w.btn_draw = btn1
        hbox.add_widget(btn1)

        btn2 = Widgets.RadioButton("Edit", group=btn1)
        btn2.set_state(mode == 'edit')
        btn2.add_callback(
            'activated', lambda w, val: self.set_mode_cb('edit', val))
        btn2.set_tooltip("Choose this to edit")
        self.w.btn_edit = btn2
        hbox.add_widget(btn2)

        hbox.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(hbox, stretch=0)

        fr = Widgets.Frame("Attributes")
        vbox2 = Widgets.VBox()
        self.w.attrlbl = Widgets.Label()
        vbox2.add_widget(self.w.attrlbl, stretch=0)
        self.w.drawvbox = Widgets.VBox()
        vbox2.add_widget(self.w.drawvbox, stretch=1)
        fr.set_widget(vbox2)

        vbox.add_widget(fr, stretch=0)

        captions = (("Rotate By:", 'label', 'Rotate By', 'entry',
                     "Scale By:", 'label', 'Scale By', 'entry'),
                    ("Delete Obj", 'button', "sp1", 'spacer',
                     "Create mask", 'button', "Clear canvas", 'button'),
                    )
        w, b = Widgets.build_info(captions)
        self.w.update(b)
        b.delete_obj.add_callback('activated', lambda w: self.delete_object())
        b.delete_obj.set_tooltip("Delete selected object in edit mode")
        b.delete_obj.set_enabled(False)
        b.scale_by.add_callback('activated', self.scale_object)
        b.scale_by.set_text('0.9')
        b.scale_by.set_tooltip("Scale selected object in edit mode")
        b.scale_by.set_enabled(False)
        b.rotate_by.add_callback('activated', self.rotate_object)
        b.rotate_by.set_text('90.0')
        b.rotate_by.set_tooltip("Rotate selected object in edit mode")
        b.rotate_by.set_enabled(False)
        b.create_mask.add_callback('activated', lambda w: self.create_mask())
        b.create_mask.set_tooltip("Create boolean mask from drawing")
        b.clear_canvas.add_callback('activated', lambda w: self.clear_canvas())
        b.clear_canvas.set_tooltip("Delete all drawing objects")

        vbox.add_widget(w, stretch=0)

        spacer = Widgets.Label('')
        vbox.add_widget(spacer, stretch=1)

        top.add_widget(sw, stretch=1)

        btns = Widgets.HBox()
        btns.set_spacing(4)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)
        top.add_widget(btns, stretch=0)

        container.add_widget(top, stretch=1)

        self.toggle_create_button()

    def close(self):
        self.fv.stop_local_plugin(self.chname, str(self))

    def start(self):
        self.set_drawparams_cb()

        # insert layer if it is not already
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.get_object_by_tag(self.layertag)

        except KeyError:
            # Add canvas layer
            p_canvas.add(self.canvas, tag=self.layertag)

        self.resume()

    def pause(self):
        self.canvas.ui_set_active(False)

    def resume(self):
        self.canvas.ui_set_active(True)
        self.fv.show_status("Draw a figure with the right mouse button")

    def stop(self):
        # remove the canvas from the image
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.delete_object_by_tag(self.layertag)
        except Exception:
            pass
        # don't leave us stuck in edit mode
        self.canvas.set_draw_mode('draw')
        self.canvas.ui_set_active(False)
        self.fv.show_status("")

    def redo(self):
        pass

    def draw_cb(self, canvas, tag):
        obj = canvas.get_object_by_tag(tag)
        self._drawn_tags.append(tag)
        self.toggle_create_button()
        self.logger.info("drew a %s" % (obj.kind))
        return True

    def set_drawtype_cb(self):
        if self.canvas.get_draw_mode() != 'draw':
            self.canvas.set_draw_mode('draw')
        self.w.btn_draw.set_state(True)
        self.set_mode_cb('draw', True)

    def set_drawparams_cb(self):
        index = self.w.draw_type.get_index()
        kind = self.drawtypes[index]

        # remove old params
        self.w.drawvbox.remove_all()

        # Create new drawing class of the right kind
        drawClass = self.canvas.get_draw_class(kind)

        self.w.attrlbl.set_text("New Object: %s" % (kind))
        # Build up a set of control widgets for the parameters
        # of the canvas object to be drawn
        paramlst = drawClass.get_params_metadata()

        params = self.drawparams_cache.setdefault(kind, Bunch.Bunch())
        self.draw_params = ParamSet.ParamSet(self.logger, params)

        w = self.draw_params.build_params(paramlst,
                                          orientation=self.orientation)
        self.draw_params.add_callback('changed', self.draw_params_changed_cb)

        self.w.drawvbox.add_widget(w, stretch=1)

        # disable edit-only controls
        self.w.delete_obj.set_enabled(False)
        self.w.scale_by.set_enabled(False)
        self.w.rotate_by.set_enabled(False)

        args, kwdargs = self.draw_params.get_params()
        self.logger.debug("changing params to: %s" % (str(kwdargs)))
        self.canvas.set_drawtype(kind, **kwdargs)

    def draw_params_changed_cb(self, paramObj, params):
        index = self.w.draw_type.get_index()
        kind = self.drawtypes[index]

        args, kwdargs = self.draw_params.get_params()
        self.logger.debug("changing params to: %s" % (str(kwdargs)))
        self.canvas.set_drawtype(kind, **kwdargs)

    def edit_cb(self, fitsimage, obj):
        # <-- obj has been edited
        #self.logger.debug("edit event on canvas: obj=%s" % (obj))
        if obj != self.edit_obj:
            # edit object is new.  Update visual parameters
            self.edit_select_cb(fitsimage, obj)
        else:
            # edit object has been modified.  Sync visual parameters
            self.draw_params.params_to_widgets()

    def edit_params_changed_cb(self, paramObj, obj):
        self.draw_params.widgets_to_params()
        if hasattr(obj, 'coord'):
            tomap = self.fitsimage.get_coordmap(obj.coord)
            if obj.crdmap != tomap:
                self.logger.debug(
                    "coordmap has changed to '%s'--"
                    "converting mapper" % (str(tomap)))
                # user changed type of mapper; convert coordinates to
                # new mapper and update widgets
                obj.convert_mapper(tomap)
                paramObj.params_to_widgets()

        obj.sync_state()
        # TODO: change whence to 0 if allowing editing of images
        whence = 2
        self.canvas.redraw(whence=whence)

    def edit_initialize(self, fitsimage, obj):
        # remove old params
        self.w.drawvbox.remove_all()

        self.edit_obj = obj
        if (obj is not None) and self.canvas.is_selected(obj):
            self.w.attrlbl.set_text("Editing a %s" % (obj.kind))

            drawClass = obj.__class__

            # Build up a set of control widgets for the parameters
            # of the canvas object to be drawn
            paramlst = drawClass.get_params_metadata()

            self.draw_params = ParamSet.ParamSet(self.logger, obj)

            w = self.draw_params.build_params(paramlst,
                                              orientation=self.orientation)
            self.draw_params.add_callback(
                'changed', self.edit_params_changed_cb)

            self.w.drawvbox.add_widget(w, stretch=1)
            self.w.delete_obj.set_enabled(True)
            self.w.scale_by.set_enabled(True)
            self.w.rotate_by.set_enabled(True)
        else:
            self.w.attrlbl.set_text("")

            self.w.delete_obj.set_enabled(False)
            self.w.scale_by.set_enabled(False)
            self.w.rotate_by.set_enabled(False)

    def edit_select_cb(self, fitsimage, obj):
        self.logger.debug(
            "editing selection status has changed for %s" % str(obj))
        self.edit_initialize(fitsimage, obj)

    def set_mode_cb(self, mode, tf):
        if tf:
            self.canvas.set_draw_mode(mode)
            if mode == 'edit':
                self.edit_initialize(self.fitsimage, None)
            elif mode == 'draw':
                self.set_drawparams_cb()
        return True

    def toggle_create_button(self):
        """Enable or disable Create Mask button based on drawn objects."""
        if len(self._drawn_tags) > 0:
            self.w.create_mask.set_enabled(True)
        else:
            self.w.create_mask.set_enabled(False)

    def create_mask(self):
        """Create boolean mask from drawing.

        All areas enclosed by all the shapes drawn will be set to 1 (True)
        in the mask. Otherwise, the values will be set to 0 (False).
        The mask will be inserted as a new image buffer, like ``Mosaic``.

        """
        ntags = len(self._drawn_tags)

        if ntags == 0:
            return

        old_image = self.fitsimage.get_image()

        if old_image is None:
            return

        mask = None
        obj_kinds = set()

        # Create mask
        for tag in self._drawn_tags:
            obj = self.canvas.get_object_by_tag(tag)

            try:
                cur_mask = old_image.get_shape_mask(obj)
            except Exception as e:
                self.logger.error('Cannot create mask: {0}'.format(str(e)))
                continue

            if mask is not None:
                mask |= cur_mask
            else:
                mask = cur_mask

            obj_kinds.add(obj.kind)

        # Might be useful to inherit header from displayed image (e.g., WCS)
        # but the displayed image should not be modified.
        # Bool needs to be converted to int so FITS writer would not crash.
        image = dp.make_image(mask.astype('int16'), old_image, {},
                              pfx=self._mask_prefix)
        imname = image.get('name')

        # Insert new image
        self.fv.gui_call(self.fv.add_image, imname, image, chname=self.chname)

        # Add description to ChangeHistory
        s = 'Mask created from {0} drawings ({1})'.format(
            ntags, ','.join(sorted(obj_kinds)))
        info = dict(time_modified=datetime.utcnow(), reason_modified=s)
        self.fv.update_image_info(image, info)
        self.logger.info(s)

    def clear_canvas(self):
        self.canvas.clear_selected()
        self.canvas.delete_all_objects()
        self._drawn_tags = []
        self.toggle_create_button()

    def delete_object(self):
        tag = self.canvas.lookup_object_tag(self.canvas._edit_obj)
        self._drawn_tags.remove(tag)
        self.toggle_create_button()
        self.canvas.edit_delete()
        self.canvas.redraw(whence=2)

    def rotate_object(self, w):
        delta = float(w.get_text())
        self.canvas.edit_rotate(delta, self.fitsimage)

    def scale_object(self, w):
        delta = float(w.get_text())
        self.canvas.edit_scale(delta, delta, self.fitsimage)

    def __str__(self):
        return 'drawing'

# END



#
# basic.py -- classes for basic shapes drawn on ginga canvases.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import math
import numpy as np

from ginga.canvas.CanvasObject import (CanvasObjectBase, _bool, _color,
                                       Point as XPoint, MovePoint, ScalePoint,
                                       EditPoint,
                                       register_canvas_types,
                                       colors_plus_none, coord_names)
from ginga import trcalc
from ginga.misc.ParamSet import Param
from ginga.util import bezier

from .mixins import (OnePointMixin, TwoPointMixin, OnePointOneRadiusMixin,
                     OnePointTwoRadiusMixin, PolygonMixin)


#
#   ==== BASIC CLASSES FOR GRAPHICS OBJECTS ====
#
class Text(OnePointMixin, CanvasObjectBase):
    """Draws text on a DrawingCanvas.
    Parameters are:
    x, y: 0-based coordinates in the data space
    text: the text to draw
    Optional parameters for fontsize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x', type=float, default=0.0, argpos=0,
                  description="X coordinate of lower left of text"),
            Param(name='y', type=float, default=0.0, argpos=1,
                  description="Y coordinate of lower left of text"),
            Param(name='text', type=str, default='EDIT ME',
                  description="Text to display"),
            Param(name='font', type=str, default='Sans Serif',
                  description="Font family for text"),
            Param(name='fontsize', type=float, default=None,
                  min=2, max=144,
                  description="Font size of text (default: vary by scale)"),
            Param(name='fontsize_min', type=float, default=6.0,
                  min=2, max=144,
                  description="Minimum font size of text (if not fixed)"),
            Param(name='fontsize_max', type=float, default=None,
                  min=2, max=144,
                  description="Maximum font size of text (if not fixed)"),
            Param(name='fontscale', type=_bool,
                  default=False, valid=[False, True],
                  description="Scale font with scale of viewer"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of text"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of text"),
            Param(name='rot_deg', type=float, default=0.0,
                  min=-359.999, max=359.999, widget='spinfloat', incr=1.0,
                  description="Rotation of text"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        return cls(cxt.start_x, cxt.start_y, **cxt.drawparams)

    def __init__(self, x, y, text='EDIT ME',
                 font='Sans Serif', fontsize=None, fontscale=False,
                 fontsize_min=6.0, fontsize_max=None,
                 color='yellow', alpha=1.0, rot_deg=0.0,
                 showcap=False, **kwdargs):
        self.kind = 'text'
        points = np.asarray([(x, y)], dtype=np.float)
        super(Text, self).__init__(points=points, color=color, alpha=alpha,
                                   font=font, fontsize=fontsize,
                                   fontscale=fontscale,
                                   fontsize_min=fontsize_min,
                                   fontsize_max=fontsize_max,
                                   text=text, rot_deg=rot_deg,
                                   showcap=showcap, **kwdargs)
        OnePointMixin.__init__(self)

    def set_edit_point(self, i, pt, detail):
        if i == 0:
            self.move_to_pt(pt)
        elif i == 1:
            scalef = self.calc_scale_from_pt(pt, detail)
            self.fontsize = detail.fontsize * scalef
        elif i == 2:
            delta_deg = self.calc_rotation_from_pt(pt, detail)
            self.rot_deg = math.fmod(self.rot_deg + delta_deg, 360.0)
        else:
            raise ValueError("No point corresponding to index %d" % (i))

    def scale_by_factors(self, factors):
        fontsize = 10.0 if self.fontsize is None else self.fontsize
        self.fontsize = fontsize * factors[0]

    def rotate_by_deg(self, thetas):
        self.rot_deg += thetas[0]

    def setup_edit(self, detail):
        detail.center_pos = self.get_center_pt()
        fontsize = self.fontsize
        if fontsize is None:
            fontsize = 10.0
        detail.fontsize = fontsize
        detail.rot_deg = self.rot_deg

    def get_edit_points(self, viewer):
        move_pt, scale_pt, rotate_pt = self.get_move_scale_rotate_pts(viewer)
        return [move_pt,
                scale_pt,
                rotate_pt,
                ]

    def select_contains_pt(self, viewer, pt):
        x1, y1, x2, y2 = self._get_unrotated_text_llur(viewer)
        # rotate point back to non-rotated cartesian alignment for test
        x, y = trcalc.rotate_pt(pt[0], pt[1], -self.rot_deg,
                                xoff=x1, yoff=y1)

        return (min(x1, x2) <= x and x <= max(x1, x2) and
                min(y1, y2) <= y and y <= max(y1, y2))

    def _get_unrotated_text_llur(self, viewer):
        # convert coordinate to data point and then pixel pt
        x1, y1 = self.get_data_points()[0]
        cx1, cy1 = viewer.get_canvas_xy(x1, y1)
        # width and height of text define bbox
        wd_px, ht_px = viewer.renderer.get_dimensions(self)
        cx2, cy2 = cx1 + wd_px, cy1 - ht_px
        # convert back to data points and construct bbox
        x2, y2 = viewer.get_data_xy(cx2, cy2)
        x1, y1, x2, y2 = self.swapxy(x1, y1, x2, y2)
        return (x1, y1, x2, y2)

    def get_llur(self):
        x, y = self.get_data_points()[0]
        r = 20
        (x1, y1), (x2, y2) = (x - r, y - r), (x + r, y + r)
        return self.swapxy(x1, y1, x2, y2)

    def draw(self, viewer):
        cr = viewer.renderer.setup_cr(self)
        cr.set_font_from_shape(self)

        x, y = self.get_data_points()[0]
        cx, cy = viewer.get_canvas_xy(x, y)
        cr.draw_text(cx, cy, self.text, rot_deg=self.rot_deg)


class Polygon(PolygonMixin, CanvasObjectBase):
    """Draws a polygon on a DrawingCanvas.
    Parameters are:
    List of (x, y) points in the polygon.  The last one is assumed to
    be connected to the first.
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            ## Param(name='points', type=list, default=[], argpos=0,
            ##       description="points making up polygon"),
            Param(name='linewidth', type=int, default=1,
                  min=1, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='fill', type=_bool,
                  default=False, valid=[False, True],
                  description="Fill the interior"),
            Param(name='fillcolor', default=None,
                  valid=colors_plus_none, type=_color,
                  description="Color of fill"),
            Param(name='fillalpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of fill"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        points = list(cxt.points)
        points.append(XPoint(cxt.x, cxt.y))
        if len(points) < 3:
            # we need at least 3 points for a polygon, so
            # revert to a line if we haven't got enough
            klass = canvas.getDrawClass('line')
            return klass(cxt.start_x, cxt.start_y, cxt.x, cxt.y,
                         **cxt.drawparams)
        else:
            return cls(points, **cxt.drawparams)

    def __init__(self, points, color='red',
                 linewidth=1, linestyle='solid', showcap=False,
                 fill=False, fillcolor=None, alpha=1.0,
                 fillalpha=1.0, **kwdargs):
        self.kind = 'polygon'

        CanvasObjectBase.__init__(self, points=points, color=color,
                                  linewidth=linewidth, showcap=showcap,
                                  linestyle=linestyle, alpha=alpha,
                                  fill=fill, fillcolor=fillcolor,
                                  fillalpha=fillalpha, **kwdargs)
        PolygonMixin.__init__(self)

        assert len(points) > 2, ValueError("Polygons need at least 3 points")

    def draw(self, viewer):
        cr = viewer.renderer.setup_cr(self)

        cpoints = self.get_cpoints(viewer)

        cr.draw_polygon(cpoints)

        if self.showcap:
            self.draw_caps(cr, self.cap, cpoints)


class Path(PolygonMixin, CanvasObjectBase):
    """Draws a path on a DrawingCanvas.
    Parameters are:
    List of (x, y) points in the polygon.
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            ## Param(name='points', type=list, default=[], argpos=0,
            ##       description="points making up polygon"),
            Param(name='linewidth', type=int, default=1,
                  min=1, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        points = list(cxt.points)
        points.append((cxt.x, cxt.y))
        return cls(points, **cxt.drawparams)

    def __init__(self, points, color='red',
                 linewidth=1, linestyle='solid', showcap=False,
                 alpha=1.0, **kwdargs):
        self.kind = 'path'

        CanvasObjectBase.__init__(self, points=points, color=color,
                                  linewidth=linewidth, showcap=showcap,
                                  linestyle=linestyle, alpha=alpha,
                                  **kwdargs)
        PolygonMixin.__init__(self)

    def contains_pts_path(self, points, path_points, radius=1.0):
        # This code is split out of contains_pts() so that it can
        # be called from BezierCurve with a different set of path
        # points.
        p_start = path_points[0]
        points = np.asarray(points)
        contains = None
        for ppt in path_points[1:]:
            p_stop = ppt
            res = self.point_within_line(points, p_start, p_stop, radius)
            if contains is None:
                contains = res
            else:
                contains = np.logical_or(contains, res)
            p_start = p_stop
        return contains

    def contains_pts(self, points, radius=1.0):
        return self.contains_pts_path(points, self.points, radius=radius)

    def select_contains_path(self, viewer, path_points, pt):
        # This code is split out so that it can be called from
        # BezierCurve with a different set of points
        p_start = path_points[0]
        for point in path_points[1:]:
            p_stop = point
            if self.within_line(viewer, pt, p_start, p_stop, self.cap_radius):
                return True
            p_start = p_stop
        return False

    def select_contains_pt(self, viewer, pt):
        path_points = self.get_data_points()
        return self.select_contains_path(viewer, path_points, pt)

    def draw(self, viewer):
        cpoints = self.get_cpoints(viewer)

        cr = viewer.renderer.setup_cr(self)
        cr.draw_path(cpoints)

        if self.showcap:
            self.draw_caps(cr, self.cap, cpoints)


class FreePolygon(Polygon):

    @classmethod
    def idraw(cls, canvas, cxt):
        cxt.points.append((cxt.x, cxt.y))
        points = list(cxt.points)
        return cls(points, **cxt.drawparams)


class FreePath(Path):

    @classmethod
    def idraw(cls, canvas, cxt):
        cxt.points.append((cxt.x, cxt.y))
        points = list(cxt.points)
        return cls(points, **cxt.drawparams)


class BezierCurve(Path):
    """Draws a Bezier Curve on a DrawingCanvas.
    Parameters are:
    List of (x, y) points in the curve.
    Optional parameters for linesize, color, etc.

    TODO: need to implement contains_pt(), which means figuring out whether a
    point lies on a bezier curve.
        See http://polymathprogrammer.com/2012/04/03/does-point-lie-on-bezier-curve/
    """

    @classmethod
    def idraw(cls, canvas, cxt):
        points = list(cxt.points)
        points.append((cxt.x, cxt.y))
        return cls(points, **cxt.drawparams)

    def __init__(self, points, color='red',
                 linewidth=1, linestyle='solid', showcap=False,
                 alpha=1.0, **kwdargs):
        self.kind = 'beziercurve'

        CanvasObjectBase.__init__(self, points=points, color=color,
                                  linewidth=linewidth, showcap=showcap,
                                  linestyle=linestyle, alpha=alpha,
                                  **kwdargs)
        PolygonMixin.__init__(self)

    def get_points_on_curve(self, image):
        points = list(self.get_data_points())
        # use maximum dimension of image to estimate a reasonable number
        # of intermediate points
        #steps = max(*image.get_size())
        steps = bezier.bezier_steps
        return list(bezier.get_4pt_bezier(steps, points))

    def select_contains_pt(self, viewer, pt):
        image = viewer.get_image()
        path_points = self.get_points_on_curve(image)
        return self.select_contains_path(viewer, path_points, pt)

    # TODO: this probably belongs somewhere else
    def get_pixels_on_curve(self, image, getvalues=True):
        data = image.get_data()
        wd, ht = image.get_size()
        if getvalues:
            res = [data[int(y), int(x)]
                   if 0 <= x < wd and 0 <= y < ht else np.NaN
                   for x, y in self.get_points_on_curve(image)]
        else:
            res = [(int(x), int(y))
                   if 0 <= x < wd and 0 <= y < ht else np.NaN
                   for x, y in self.get_points_on_curve(image)]
        return res

    def draw(self, viewer):
        cpoints = self.get_cpoints(viewer)

        cr = viewer.renderer.setup_cr(self)
        if len(cpoints) < 4:
            # until we have 4 points, we cannot draw a quadradic bezier curve
            cr.draw_path(cpoints)
        else:
            if hasattr(cr, 'draw_bezier_curve'):
                cr.draw_bezier_curve(cpoints)

            else:
                # No Bezier support in this backend, so calculate intermediate
                # points and draw a path
                #steps = max(*viewer.get_window_size())
                steps = bezier.bezier_steps
                ipoints = list(bezier.get_4pt_bezier(steps, cpoints))
                cr.draw_path(ipoints)

        if self.showcap:
            self.draw_caps(cr, self.cap, cpoints)


class Box(OnePointTwoRadiusMixin, CanvasObjectBase):
    """Draws a box on a DrawingCanvas.
    Parameters are:
    x, y: 0-based coordinates of the center in the data space
    xradius, yradius: radii based on the number of pixels in data space
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x', type=float, default=0.0, argpos=0,
                  description="X coordinate of center of object"),
            Param(name='y', type=float, default=0.0, argpos=1,
                  description="Y coordinate of center of object"),
            Param(name='xradius', type=float, default=1.0, argpos=2,
                  min=0.0,
                  description="X radius of object"),
            Param(name='yradius', type=float, default=1.0, argpos=3,
                  min=0.0,
                  description="Y radius of object"),
            Param(name='linewidth', type=int, default=1,
                  min=1, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='fill', type=_bool,
                  default=False, valid=[False, True],
                  description="Fill the interior"),
            Param(name='fillcolor', default=None,
                  valid=colors_plus_none, type=_color,
                  description="Color of fill"),
            Param(name='fillalpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of fill"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
            Param(name='rot_deg', type=float, default=0.0,
                  min=-359.999, max=359.999, widget='spinfloat', incr=1.0,
                  description="Rotation about center of object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        xradius, yradius = abs(cxt.start_x - cxt.x), abs(cxt.start_y - cxt.y)
        return cls(cxt.start_x, cxt.start_y, xradius, yradius, **cxt.drawparams)

    def __init__(self, x, y, xradius, yradius, color='red',
                 linewidth=1, linestyle='solid', showcap=False,
                 fill=False, fillcolor=None, alpha=1.0, fillalpha=1.0,
                 rot_deg=0.0, **kwdargs):
        points = np.asarray([(x, y)], dtype=np.float)
        CanvasObjectBase.__init__(self, points=points, color=color,
                                  linewidth=linewidth, showcap=showcap,
                                  linestyle=linestyle,
                                  fill=fill, fillcolor=fillcolor,
                                  alpha=alpha, fillalpha=fillalpha,
                                  xradius=xradius, yradius=yradius,
                                  rot_deg=rot_deg,
                                  **kwdargs)
        OnePointTwoRadiusMixin.__init__(self)
        self.kind = 'box'

    def get_points(self):
        points = (self.crdmap.offset_pt((self.x, self.y),
                                        (-self.xradius, -self.yradius)),
                  self.crdmap.offset_pt((self.x, self.y),
                                        (self.xradius, -self.yradius)),
                  self.crdmap.offset_pt((self.x, self.y),
                                        (self.xradius, self.yradius)),
                  self.crdmap.offset_pt((self.x, self.y),
                                        (-self.xradius, self.yradius)),
                  )
        points = self.get_data_points(points=points)
        return points

    def contains_pts(self, pts):
        x_arr, y_arr = np.asarray(pts).T

        points = self.get_points()
        x1, y1 = points[0]
        x2, y2 = points[2]

        # rotate points back to non-rotated cartesian alignment for test
        xd, yd = self.crdmap.to_data((self.x, self.y))
        xa, ya = trcalc.rotate_pt(x_arr, y_arr, -self.rot_deg,
                                  xoff=xd, yoff=yd)

        contains = np.logical_and(
            np.logical_and(min(x1, x2) <= xa, xa <= max(x1, x2)),
            np.logical_and(min(y1, y2) <= ya, ya <= max(y1, y2)))
        return contains

    def draw(self, viewer):
        cpoints = self.get_cpoints(viewer)

        cr = viewer.renderer.setup_cr(self)
        cr.draw_polygon(cpoints)

        if self.showcap:
            self.draw_caps(cr, self.cap, cpoints)


class SquareBox(OnePointOneRadiusMixin, CanvasObjectBase):
    """Draws a square box on a DrawingCanvas.
    Parameters are:
    x, y: 0-based coordinates of the center in the data space
    radius: radius based on the number of pixels in data space
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x', type=float, default=0.0, argpos=0,
                  description="X coordinate of center of object"),
            Param(name='y', type=float, default=0.0, argpos=1,
                  description="Y coordinate of center of object"),
            Param(name='radius', type=float, default=1.0, argpos=2,
                  min=0.0,
                  description="radius of object"),
            Param(name='linewidth', type=int, default=1,
                  min=1, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='fill', type=_bool,
                  default=False, valid=[False, True],
                  description="Fill the interior"),
            Param(name='fillcolor', default=None,
                  valid=colors_plus_none, type=_color,
                  description="Color of fill"),
            Param(name='fillalpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of fill"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
            Param(name='rot_deg', type=float, default=0.0,
                  min=-359.999, max=359.999, widget='spinfloat', incr=1.0,
                  description="Rotation about center of object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        len_x = cxt.start_x - cxt.x
        len_y = cxt.start_y - cxt.y
        radius = max(abs(len_x), abs(len_y))
        return cls(cxt.start_x, cxt.start_y, radius, **cxt.drawparams)

    def __init__(self, x, y, radius, color='red',
                 linewidth=1, linestyle='solid', showcap=False,
                 fill=False, fillcolor=None, alpha=1.0, fillalpha=1.0,
                 rot_deg=0.0, **kwdargs):
        points = np.asarray([(x, y)], dtype=np.float)
        CanvasObjectBase.__init__(self, points=points, color=color,
                                  linewidth=linewidth, showcap=showcap,
                                  linestyle=linestyle,
                                  fill=fill, fillcolor=fillcolor,
                                  alpha=alpha, fillalpha=fillalpha,
                                  radius=radius, rot_deg=rot_deg,
                                  **kwdargs)
        OnePointOneRadiusMixin.__init__(self)
        self.kind = 'squarebox'

    def get_points(self):
        points = (self.crdmap.offset_pt((self.x, self.y),
                                        (-self.radius, -self.radius)),
                  self.crdmap.offset_pt((self.x, self.y),
                                        (self.radius, -self.radius)),
                  self.crdmap.offset_pt((self.x, self.y),
                                        (self.radius, self.radius)),
                  self.crdmap.offset_pt((self.x, self.y),
                                        (-self.radius, self.radius)))
        points = self.get_data_points(points=points)
        return points

    def rotate_by_deg(self, thetas):
        new_rot = np.fmod(self.rot_deg + thetas[0], 360.0)
        self.rot_deg = new_rot
        return new_rot

    def contains_pts(self, pts):
        x_arr, y_arr = np.asarray(pts).T

        points = self.get_points()
        x1, y1 = points[0]
        x2, y2 = points[2]

        # rotate point back to cartesian alignment for test
        xd, yd = self.crdmap.to_data((self.x, self.y))
        xa, ya = trcalc.rotate_pt(x_arr, y_arr, -self.rot_deg,
                                  xoff=xd, yoff=yd)

        contains = np.logical_and(
            np.logical_and(min(x1, x2) <= xa, xa <= max(x1, x2)),
            np.logical_and(min(y1, y2) <= ya, ya <= max(y1, y2)))
        return contains

    def set_edit_point(self, i, pt, detail):
        if i == 0:
            self.set_point_by_index(i, pt)
        elif i == 1:
            scalef = self.calc_scale_from_pt(pt, detail)
            self.radius = detail.radius * scalef
        elif i == 2:
            delta_deg = self.calc_rotation_from_pt(pt, detail)
            self.rotate_by_deg([delta_deg])
        else:
            raise ValueError("No point corresponding to index %d" % (i))

    def get_edit_points(self, viewer):
        points = self.get_data_points(points=(
            self.crdmap.offset_pt((self.x, self.y),
                                  (self.radius, self.radius)),
        ))
        move_pt, scale_pt, rotate_pt = self.get_move_scale_rotate_pts(viewer)
        return [move_pt,
                ScalePoint(*points[0]),
                rotate_pt,
                ]

    def draw(self, viewer):
        cpoints = self.get_cpoints(viewer)

        cr = viewer.renderer.setup_cr(self)
        cr.draw_polygon(cpoints)

        if self.showcap:
            self.draw_caps(cr, self.cap, cpoints)


class Ellipse(OnePointTwoRadiusMixin, CanvasObjectBase):
    """Draws an ellipse on a DrawingCanvas.
    Parameters are:
    x, y: 0-based coordinates of the center in the data space
    xradius, yradius: radii based on the number of pixels in data space
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x', type=float, default=0.0, argpos=0,
                  description="X coordinate of center of object"),
            Param(name='y', type=float, default=0.0, argpos=1,
                  description="Y coordinate of center of object"),
            Param(name='xradius', type=float, default=1.0, argpos=2,
                  min=0.0,
                  description="X radius of object"),
            Param(name='yradius', type=float, default=1.0, argpos=3,
                  min=0.0,
                  description="Y radius of object"),
            Param(name='linewidth', type=int, default=1,
                  min=1, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='fill', type=_bool,
                  default=False, valid=[False, True],
                  description="Fill the interior"),
            Param(name='fillcolor', default=None,
                  valid=colors_plus_none, type=_color,
                  description="Color of fill"),
            Param(name='fillalpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of fill"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
            Param(name='rot_deg', type=float, default=0.0,
                  min=-359.999, max=359.999, widget='spinfloat', incr=1.0,
                  description="Rotation about center of object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        xradius, yradius = abs(cxt.start_x - cxt.x), abs(cxt.start_y - cxt.y)
        return cls(cxt.start_x, cxt.start_y, xradius, yradius, **cxt.drawparams)

    def __init__(self, x, y, xradius, yradius, color='yellow',
                 linewidth=1, linestyle='solid', showcap=False,
                 fill=False, fillcolor=None, alpha=1.0, fillalpha=1.0,
                 rot_deg=0.0, **kwdargs):
        points = np.asarray([(x, y)], dtype=np.float)
        CanvasObjectBase.__init__(self, points=points, color=color,
                                  linewidth=linewidth, showcap=showcap,
                                  linestyle=linestyle,
                                  fill=fill, fillcolor=fillcolor,
                                  alpha=alpha, fillalpha=fillalpha,
                                  xradius=xradius, yradius=yradius,
                                  rot_deg=rot_deg, **kwdargs)
        OnePointTwoRadiusMixin.__init__(self)
        self.kind = 'ellipse'

    def get_points(self):
        points = ((self.x, self.y),
                  self.crdmap.offset_pt((self.x, self.y),
                                        (self.xradius, 0)),
                  self.crdmap.offset_pt((self.x, self.y),
                                        (0, self.yradius)),
                  self.crdmap.offset_pt((self.x, self.y),
                                        (self.xradius, self.yradius)),
                  )
        points = self.get_data_points(points=points)
        return points

    def contains_pts(self, pts):
        x_arr, y_arr = np.asarray(pts).T
        x_arr, y_arr = (x_arr.astype(np.float, copy=False),
                        y_arr.astype(np.float, copy=False))

        points = self.get_points()
        # rotate point back to cartesian alignment for test
        xd, yd = points[0]
        xa, ya = trcalc.rotate_pt(x_arr, y_arr, -self.rot_deg,
                                  xoff=xd, yoff=yd)

        # need to recalculate radius in case of wcs coords
        x2, y2 = points[3]
        xradius = max(x2, xd) - min(x2, xd)
        yradius = max(y2, yd) - min(y2, yd)

        # See http://math.stackexchange.com/questions/76457/check-if-a-point-is-within-an-ellipse
        res = (((xa - xd) ** 2) / xradius ** 2 +
               ((ya - yd) ** 2) / yradius ** 2)
        contains = (res <= 1.0)
        return contains

    def get_wdht(self):
        # See http://stackoverflow.com/questions/87734/how-do-you-calculate-the-axis-aligned-bounding-box-of-an-ellipse
        theta = np.radians(self.rot_deg)
        sin_theta = np.sin(theta)
        cos_theta = np.cos(theta)

        # need to recalculate radius in case of wcs coords
        points = self.get_points()
        x, y = points[0]
        xradius = abs(points[1][0] - x)
        yradius = abs(points[2][1] - y)

        a = xradius * cos_theta
        b = yradius * sin_theta
        c = xradius * sin_theta
        d = yradius * cos_theta
        wd = np.sqrt(a ** 2.0 + b ** 2.0) * 2.
        ht = np.sqrt(c ** 2.0 + d ** 2.0) * 2.

        return (points, (wd, ht))

    def get_llur(self):
        points, dims = self.get_wdht()
        x, y = points[0]
        wd, ht = dims

        x1, y1 = x - wd * 0.5, y + ht * 0.5
        x2, y2 = x1 + wd, y1 - ht
        return self.swapxy(x1, y1, x2, y2)

    def get_bezier_pts(self, points, rot_deg=None):
        x, y = points[0]
        xradius = abs(points[1][0] - x)
        yradius = abs(points[2][1] - y)

        pts = np.asarray(bezier.get_bezier_ellipse(x, y, xradius, yradius))
        if rot_deg is None:
            return pts

        # specified a rotation for the points
        return trcalc.rotate_coord(pts, [rot_deg], (x, y))

    def draw(self, viewer):
        cr = viewer.renderer.setup_cr(self)

        if hasattr(cr, 'draw_ellipse'):
            # <- backend can draw rotated ellipses
            cpoints = self.get_cpoints(viewer,
                                       points=self.get_edit_points(viewer))
            cx, cy = cpoints[0]
            cxradius = abs(cpoints[1][0] - cx)
            cyradius = abs(cpoints[2][1] - cy)
            cr.draw_ellipse(cx, cy, cxradius, cyradius, self.rot_deg)

        elif hasattr(cr, 'draw_ellipse_bezier'):
            # <- backend can draw Bezier curves
            points = self.get_bezier_pts(self.get_points())
            cp = self.get_cpoints(viewer, points=points)
            cr.draw_ellipse_bezier(cp)

        else:
            # <- backend can draw polygons
            points = self.get_bezier_pts(self.get_points())
            cp = self.get_cpoints(viewer, points=points)
            num_pts = bezier.bezier_steps
            cp = bezier.get_bezier(num_pts, cp)
            cr.draw_polygon(cp)

        if self.showcap:
            cpoints = self.get_cpoints(viewer)
            self.draw_caps(cr, self.cap, cpoints)


class Triangle(OnePointTwoRadiusMixin, CanvasObjectBase):
    """Draws a triangle on a DrawingCanvas.
    Parameters are:
    x, y: 0-based coordinates of the center in the data space
    xradius, yradius: radii based on the number of pixels in data space
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x', type=float, default=0.0, argpos=0,
                  description="X coordinate of center of object"),
            Param(name='y', type=float, default=0.0, argpos=1,
                  description="Y coordinate of center of object"),
            Param(name='xradius', type=float, default=1.0, argpos=2,
                  min=0.0,
                  description="X radius of object"),
            Param(name='yradius', type=float, default=1.0, argpos=3,
                  min=0.0,
                  description="Y radius of object"),
            Param(name='linewidth', type=int, default=1,
                  min=1, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='fill', type=_bool,
                  default=False, valid=[False, True],
                  description="Fill the interior"),
            Param(name='fillcolor', default=None,
                  valid=colors_plus_none, type=_color,
                  description="Color of fill"),
            Param(name='fillalpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of fill"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
            Param(name='rot_deg', type=float, default=0.0,
                  min=-359.999, max=359.999, widget='spinfloat', incr=1.0,
                  description="Rotation about center of object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        xradius, yradius = abs(cxt.start_x - cxt.x), abs(cxt.start_y - cxt.y)
        return cls(cxt.start_x, cxt.start_y, xradius, yradius, **cxt.drawparams)

    def __init__(self, x, y, xradius, yradius, color='pink',
                 linewidth=1, linestyle='solid', showcap=False,
                 fill=False, fillcolor=None, alpha=1.0, fillalpha=1.0,
                 rot_deg=0.0, **kwdargs):
        self.kind = 'triangle'
        points = np.asarray([(x, y)], dtype=np.float)
        CanvasObjectBase.__init__(self, points=points, color=color,
                                  linewidth=linewidth, showcap=showcap,
                                  linestyle=linestyle, alpha=alpha,
                                  fill=fill, fillcolor=fillcolor,
                                  fillalpha=fillalpha,
                                  xradius=xradius, yradius=yradius,
                                  rot_deg=rot_deg, **kwdargs)
        OnePointTwoRadiusMixin.__init__(self)

    def get_points(self):
        points = (self.crdmap.offset_pt((self.x, self.y),
                                        (-2 * self.xradius, -self.yradius)),
                  self.crdmap.offset_pt((self.x, self.y),
                                        (2 * self.xradius, -self.yradius)),
                  self.crdmap.offset_pt((self.x, self.y),
                                        (0, self.yradius)),
                  )
        points = self.get_data_points(points=points)
        return points

    def get_llur(self):
        xd, yd = self.crdmap.to_data((self.x, self.y))
        points = np.asarray(self.get_points(), dtype=np.float)

        mpts = trcalc.rotate_coord(points, [self.rot_deg], [xd, yd])
        t_ = mpts.T

        x1, y1 = t_[0].min(), t_[1].min()
        x2, y2 = t_[0].max(), t_[1].max()
        return (x1, y1, x2, y2)

    def contains_pts(self, pts):
        x_arr, y_arr = np.asarray(pts).T
        x_arr, y_arr = (x_arr.astype(np.float, copy=False),
                        y_arr.astype(np.float, copy=False))
        # is this the same as self.x, self.y ?
        xd, yd = self.get_center_pt()
        # rotate point back to cartesian alignment for test
        xa, ya = trcalc.rotate_pt(x_arr, y_arr, -self.rot_deg,
                                  xoff=xd, yoff=yd)

        (x1, y1), (x2, y2), (x3, y3) = self.get_points()

        # barycentric coordinate test
        denominator = float((y2 - y3) * (x1 - x3) + (x3 - x2) * (y1 - y3))
        a = ((y2 - y3) * (xa - x3) + (x3 - x2) * (ya - y3)) / denominator
        b = ((y3 - y1) * (xa - x3) + (x1 - x3) * (ya - y3)) / denominator
        c = 1.0 - a - b

        #tf = (0.0 <= a <= 1.0 and 0.0 <= b <= 1.0 and 0.0 <= c <= 1.0)
        contains = np.logical_and(
            np.logical_and(0.0 <= a, a <= 1.0),
            np.logical_and(np.logical_and(0.0 <= b, b <= 1.0),
                           np.logical_and(0.0 <= c, c <= 1.0)))
        return contains

    def draw(self, viewer):
        cpoints = self.get_cpoints(viewer)

        cr = viewer.renderer.setup_cr(self)
        cr.draw_polygon(cpoints)

        if self.showcap:
            self.draw_caps(cr, self.cap, cpoints)


class Circle(OnePointOneRadiusMixin, CanvasObjectBase):
    """Draws a circle on a DrawingCanvas.
    Parameters are:
    x, y: 0-based coordinates of the center in the data space
    radius: radius based on the number of pixels in data space
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x', type=float, default=0.0, argpos=0,
                  description="X coordinate of center of object"),
            Param(name='y', type=float, default=0.0, argpos=1,
                  description="Y coordinate of center of object"),
            Param(name='radius', type=float, default=1.0, argpos=2,
                  min=0.0,
                  description="Radius of object"),
            Param(name='linewidth', type=int, default=1,
                  min=1, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='fill', type=_bool,
                  default=False, valid=[False, True],
                  description="Fill the interior"),
            Param(name='fillcolor', default=None,
                  valid=colors_plus_none, type=_color,
                  description="Color of fill"),
            Param(name='fillalpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of fill"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        radius = np.sqrt(abs(cxt.start_x - cxt.x) ** 2 +
                         abs(cxt.start_y - cxt.y) ** 2)
        return cls(cxt.start_x, cxt.start_y, radius, **cxt.drawparams)

    def __init__(self, x, y, radius, color='yellow',
                 linewidth=1, linestyle='solid', showcap=False,
                 fill=False, fillcolor=None, alpha=1.0, fillalpha=1.0,
                 **kwdargs):
        points = np.asarray([(x, y)], dtype=np.float)
        CanvasObjectBase.__init__(self, points=points, color=color,
                                  linewidth=linewidth, showcap=showcap,
                                  linestyle=linestyle,
                                  fill=fill, fillcolor=fillcolor,
                                  alpha=alpha, fillalpha=fillalpha,
                                  radius=radius, **kwdargs)
        OnePointOneRadiusMixin.__init__(self)
        self.kind = 'circle'

    def contains_pts(self, pts):
        x_arr, y_arr = np.asarray(pts).T
        x_arr, y_arr = (x_arr.astype(np.float, copy=False),
                        y_arr.astype(np.float, copy=False))

        xd, yd = self.crdmap.to_data((self.x, self.y))

        # need to recalculate radius in case of wcs coords
        points = self.get_data_points(points=(
            self.crdmap.offset_pt((self.x, self.y), (self.radius, 0)),
            self.crdmap.offset_pt((self.x, self.y), (0, self.radius)),
        ))
        (x2, y2), (x3, y3) = points
        xradius = max(x2, xd) - min(x2, xd)
        yradius = max(y3, yd) - min(y3, yd)

        # See http://math.stackexchange.com/questions/76457/check-if-a-point-is-within-an-ellipse
        res = (((x_arr - xd) ** 2) / xradius ** 2 +
               ((y_arr - yd) ** 2) / yradius ** 2)
        contains = (res <= 1.0)
        return contains

    def get_edit_points(self, viewer):
        points = self.get_data_points(points=(
            (self.x, self.y),
            self.crdmap.offset_pt((self.x, self.y), (self.radius, 0)),
        ))
        return [MovePoint(*points[0]),
                ScalePoint(*points[1]),
                ]

    def get_llur(self):
        points = self.get_data_points(points=(
            self.crdmap.offset_pt((self.x, self.y),
                                  (-self.radius, -self.radius)),
            self.crdmap.offset_pt((self.x, self.y),
                                  (self.radius, self.radius)),
        ))
        (x1, y1), (x2, y2) = points
        return self.swapxy(x1, y1, x2, y2)

    def draw(self, viewer):
        points = self.get_data_points(points=(
            (self.x, self.y),
            self.crdmap.offset_pt((self.x, self.y), (0, self.radius)),
        ))
        cpoints = self.get_cpoints(viewer, points=points)
        cx, cy, cradius = self.calc_radius(viewer,
                                           cpoints[0], cpoints[1])
        cr = viewer.renderer.setup_cr(self)
        cr.draw_circle(cx, cy, cradius)

        if self.showcap:
            self.draw_caps(cr, self.cap, ((cx, cy), ))


class Point(OnePointOneRadiusMixin, CanvasObjectBase):
    """Draws a point on a DrawingCanvas.
    Parameters are:
    x, y: 0-based coordinates of the center in the data space
    radius: radius based on the number of pixels in data space
    Optional parameters for linesize, color, style, etc.
    Currently the only styles are 'cross' and 'plus'.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x', type=float, default=0.0, argpos=0,
                  description="X coordinate of center of object"),
            Param(name='y', type=float, default=0.0, argpos=1,
                  description="Y coordinate of center of object"),
            Param(name='radius', type=float, default=1.0, argpos=2,
                  min=0.0,
                  description="Radius of object"),
            Param(name='style', type=str, default='cross',
                  valid=['cross', 'plus', 'circle', 'square', 'diamond',
                         'hexagon', 'downtriangle', 'uptriangle'],
                  description="Style of point (default 'cross')"),
            Param(name='linewidth', type=int, default=1,
                  min=1, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        radius = max(abs(cxt.start_x - cxt.x),
                     abs(cxt.start_y - cxt.y))
        return cls(cxt.start_x, cxt.start_y, radius, **cxt.drawparams)

    def __init__(self, x, y, radius, style='cross', color='yellow',
                 linewidth=1, linestyle='solid', alpha=1.0, showcap=False,
                 **kwdargs):
        self.kind = 'point'
        points = np.asarray([(x, y)], dtype=np.float)
        CanvasObjectBase.__init__(self, points=points, color=color,
                                  linewidth=linewidth, alpha=alpha,
                                  linestyle=linestyle, radius=radius,
                                  showcap=showcap, style=style,
                                  **kwdargs)
        OnePointOneRadiusMixin.__init__(self)

    def contains_pts(self, pts, radius=1.0):
        points = np.asarray(pts)
        pt = self.crdmap.to_data((self.x, self.y))
        contains = self.point_within_radius(points, pt, radius)
        return contains

    def select_contains_pt(self, viewer, pt):
        p0 = self.get_data_points()[0]
        scale = viewer.get_scale_max()
        return self.within_radius(viewer, pt, p0, self.radius * scale)

    def get_edit_points(self, viewer):
        points = self.get_data_points(points=[(self.x, self.y)])
        move_pt, scale_pt, rotate_pt = self.get_move_scale_rotate_pts(viewer)
        return [MovePoint(*points[0]), scale_pt]

    def draw(self, viewer):
        points = self.get_data_points(points=(
            (self.x, self.y),
            self.crdmap.offset_pt((self.x, self.y), (0, self.radius)),
        ))
        cpoints = self.get_cpoints(viewer, points=points)
        cx, cy, cradius = self.calc_radius(viewer,
                                           cpoints[0], cpoints[1])

        cr = viewer.renderer.setup_cr(self)

        cx1, cy1 = cx - cradius, cy - cradius
        cx2, cy2 = cx + cradius, cy + cradius

        if self.style == 'cross':
            cr.draw_line(cx1, cy1, cx2, cy2)
            cr.draw_line(cx1, cy2, cx2, cy1)

        elif self.style == 'plus':
            cr.draw_line(cx1, cy, cx2, cy)
            cr.draw_line(cx, cy1, cx, cy2)

        elif self.style == 'circle':
            cr.draw_circle(cx, cy, cradius)

        elif self.style == 'square':
            cpts = [(cx1, cy1), (cx2, cy1), (cx2, cy2), (cx1, cy2)]
            cr.draw_polygon(cpts)

        elif self.style == 'diamond':
            cpts = [(cx, cy1), ((cx + cx2) / 2.0, cy),
                    (cx, cy2), ((cx1 + cx) / 2.0, cy)]
            cr.draw_polygon(cpts)

        elif self.style == 'hexagon':
            cpts = [(cx1, cy), ((cx1 + cx) / 2.0, cy2), ((cx + cx2) / 2.0, cy2),
                    (cx2, cy), ((cx + cx2) / 2.0, cy1), ((cx1 + cx) / 2.0, cy1)]
            cr.draw_polygon(cpts)

        elif self.style == 'downtriangle':
            cpts = [(cx1, cy1), (cx2, cy1), (cx, cy2)]
            cr.draw_polygon(cpts)

        elif self.style == 'uptriangle':
            cpts = [(cx1, cy2), (cx2, cy2), (cx, cy1)]
            cr.draw_polygon(cpts)

        else:
            raise ValueError("Don't understand draw style '{}' of point".format(self.style))

        if self.showcap:
            self.draw_caps(cr, self.cap, ((cx, cy), ))


class Rectangle(TwoPointMixin, CanvasObjectBase):
    """Draws a rectangle on a DrawingCanvas.
    Parameters are:
    x1, y1: 0-based coordinates of one corner in the data space
    x2, y2: 0-based coordinates of the opposing corner in the data space
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=['data', 'wcs', 'cartesian', 'window'],
                  description="Set type of coordinates"),
            Param(name='x1', type=float, default=0.0, argpos=0,
                  description="First X coordinate of object"),
            Param(name='y1', type=float, default=0.0, argpos=1,
                  description="First Y coordinate of object"),
            Param(name='x2', type=float, default=0.0, argpos=2,
                  description="Second X coordinate of object"),
            Param(name='y2', type=float, default=0.0, argpos=3,
                  description="Second Y coordinate of object"),
            Param(name='linewidth', type=int, default=1,
                  min=1, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='fill', type=_bool,
                  default=False, valid=[False, True],
                  description="Fill the interior"),
            Param(name='fillcolor', default=None,
                  valid=colors_plus_none, type=_color,
                  description="Color of fill"),
            Param(name='fillalpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of fill"),
            Param(name='drawdims', type=_bool,
                  default=False, valid=[False, True],
                  description="Annotate with dimensions of object"),
            Param(name='font', type=str, default='Sans Serif',
                  description="Font family for text"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        return cls(cxt.start_x, cxt.start_y, cxt.x, cxt.y, **cxt.drawparams)

    def __init__(self, x1, y1, x2, y2, color='red',
                 linewidth=1, linestyle='solid', showcap=False,
                 fill=False, fillcolor=None, alpha=1.0,
                 drawdims=False, font='Sans Serif', fillalpha=1.0,
                 **kwdargs):
        self.kind = 'rectangle'
        points = np.asarray([(x1, y1), (x2, y2)], dtype=np.float)

        CanvasObjectBase.__init__(self, points=points, color=color,
                                  linewidth=linewidth, showcap=showcap,
                                  linestyle=linestyle,
                                  fill=fill, fillcolor=fillcolor,
                                  alpha=alpha, fillalpha=fillalpha,
                                  drawdims=drawdims, font=font,
                                  **kwdargs)
        TwoPointMixin.__init__(self)

    def get_points(self):
        points = [(self.x1, self.y1), (self.x2, self.y1),
                  (self.x2, self.y2), (self.x1, self.y2)]
        points = self.get_data_points(points=points)
        return points

    def contains_pts(self, pts):
        x_arr, y_arr = np.asarray(pts).T
        x1, y1, x2, y2 = self.get_llur()

        contains = np.logical_and(
            np.logical_and(x1 <= x_arr, x_arr <= x2),
            np.logical_and(y1 <= y_arr, y_arr <= y2))
        return contains

    # TO BE DEPRECATED?
    def move_point(self):
        return self.get_center_pt()

    def draw(self, viewer):
        cr = viewer.renderer.setup_cr(self)

        cpoints = self.get_cpoints(viewer)
        cr.draw_polygon(cpoints)

        if self.drawdims:
            fontsize = self.scale_font(viewer)
            cr.set_font(self.font, fontsize, color=self.color)

            cx1, cy1 = cpoints[0]
            cx2, cy2 = cpoints[2]

            # draw label on X dimension
            cx = cx1 + (cx2 - cx1) // 2
            cy = cy2 + -4
            cr.draw_text(cx, cy, "%f" % abs(self.x2 - self.x1))

            # draw label on Y dimension
            cy = cy1 + (cy2 - cy1) // 2
            cx = cx2 + 4
            cr.draw_text(cx, cy, "%f" % abs(self.y2 - self.y1))

        if self.showcap:
            self.draw_caps(cr, self.cap, cpoints)


class Square(Rectangle):

    @classmethod
    def idraw(cls, canvas, cxt):
        len_x = cxt.start_x - cxt.x
        len_y = cxt.start_y - cxt.y
        length = max(abs(len_x), abs(len_y))
        len_x = np.sign(len_x) * length
        len_y = np.sign(len_y) * length
        return cls(cxt.start_x, cxt.start_y,
                   cxt.start_x - len_x, cxt.start_y - len_y,
                   **cxt.drawparams)


class Line(TwoPointMixin, CanvasObjectBase):
    """Draws a line on a DrawingCanvas.
    Parameters are:
    x1, y1: 0-based coordinates of one end in the data space
    x2, y2: 0-based coordinates of the opposing end in the data space
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x1', type=float, default=0.0, argpos=0,
                  description="First X coordinate of object"),
            Param(name='y1', type=float, default=0.0, argpos=1,
                  description="First Y coordinate of object"),
            Param(name='x2', type=float, default=0.0, argpos=2,
                  description="Second X coordinate of object"),
            Param(name='y2', type=float, default=0.0, argpos=3,
                  description="Second Y coordinate of object"),
            Param(name='linewidth', type=int, default=1,
                  min=1, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='arrow', type=str, default='none',
                  valid=['start', 'end', 'both', 'none'],
                  description="Arrows at ends (default: none)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        return cls(cxt.start_x, cxt.start_y, cxt.x, cxt.y, **cxt.drawparams)

    def __init__(self, x1, y1, x2, y2, color='red',
                 linewidth=1, linestyle='solid', alpha=1.0,
                 arrow=None, showcap=False, **kwdargs):
        self.kind = 'line'
        points = np.asarray([(x1, y1), (x2, y2)], dtype=np.float)
        CanvasObjectBase.__init__(self, points=points, color=color, alpha=alpha,
                                  linewidth=linewidth, showcap=showcap,
                                  linestyle=linestyle, arrow=arrow,
                                  **kwdargs)
        TwoPointMixin.__init__(self)

    def get_points(self):
        points = [(self.x1, self.y1), (self.x2, self.y2)]
        points = self.get_data_points(points=points)
        return points

    def contains_pts(self, pts, radius=1.0):
        points = self.get_points()
        contains = self.point_within_line(pts, points[0], points[1],
                                          radius)
        return contains

    def select_contains_pt(self, viewer, pt):
        points = self.get_points()
        return self.within_line(viewer, pt, points[0], points[1],
                                self.cap_radius)

    def draw(self, viewer):
        points = self.get_points()
        x1, y1 = points[0]
        x2, y2 = points[1]
        cx1, cy1 = viewer.get_canvas_xy(x1, y1)
        cx2, cy2 = viewer.get_canvas_xy(x2, y2)

        cr = viewer.renderer.setup_cr(self)
        cr.draw_line(cx1, cy1, cx2, cy2)

        if self.arrow == 'end':
            self.draw_arrowhead(cr, cx1, cy1, cx2, cy2)
            caps = [(cx1, cy1)]
        elif self.arrow == 'start':
            self.draw_arrowhead(cr, cx2, cy2, cx1, cy1)
            caps = [(cx2, cy2)]
        elif self.arrow == 'both':
            self.draw_arrowhead(cr, cx2, cy2, cx1, cy1)
            self.draw_arrowhead(cr, cx1, cy1, cx2, cy2)
            caps = []
        else:
            caps = [(cx1, cy1), (cx2, cy2)]

        if self.showcap:
            self.draw_caps(cr, self.cap, caps)


class RightTriangle(TwoPointMixin, CanvasObjectBase):
    """Draws a right triangle on a DrawingCanvas.
    Parameters are:
    x1, y1: 0-based coordinates of one end of the diagonal in the data space
    x2, y2: 0-based coordinates of the opposite end of the diagonal
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x1', type=float, default=0.0, argpos=0,
                  description="First X coordinate of object"),
            Param(name='y1', type=float, default=0.0, argpos=1,
                  description="First Y coordinate of object"),
            Param(name='x2', type=float, default=0.0, argpos=2,
                  description="Second X coordinate of object"),
            Param(name='y2', type=float, default=0.0, argpos=3,
                  description="Second Y coordinate of object"),
            Param(name='linewidth', type=int, default=1,
                  min=1, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='fill', type=_bool,
                  default=False, valid=[False, True],
                  description="Fill the interior"),
            Param(name='fillcolor', default=None,
                  valid=colors_plus_none, type=_color,
                  description="Color of fill"),
            Param(name='fillalpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of fill"),
            Param(name='showcap', type=_bool,
                  default=False, valid=[False, True],
                  description="Show caps for this object"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        return cls(cxt.start_x, cxt.start_y, cxt.x, cxt.y, **cxt.drawparams)

    def __init__(self, x1, y1, x2, y2, color='pink',
                 linewidth=1, linestyle='solid', showcap=False,
                 fill=False, fillcolor=None, alpha=1.0, fillalpha=1.0,
                 **kwdargs):
        self.kind = 'righttriangle'
        points = np.asarray([(x1, y1), (x2, y2)], dtype=np.float)
        CanvasObjectBase.__init__(self, points=points, color=color, alpha=alpha,
                                  linewidth=linewidth, showcap=showcap,
                                  linestyle=linestyle,
                                  fill=fill, fillcolor=fillcolor,
                                  fillalpha=fillalpha,
                                  **kwdargs)
        TwoPointMixin.__init__(self)

    def get_points(self):
        points = [(self.x1, self.y1), (self.x2, self.y2),
                  (self.x2, self.y1)]
        points = self.get_data_points(points=points)
        return points

    def contains_pts(self, pts):
        x_arr, y_arr = np.asarray(pts).T
        x_arr, y_arr = (x_arr.astype(np.float, copy=False),
                        y_arr.astype(np.float, copy=False))
        points = self.get_points()
        x1, y1 = points[0]
        x2, y2 = points[1]
        x3, y3 = points[2]

        # barycentric coordinate test
        denominator = float((y2 - y3) * (x1 - x3) + (x3 - x2) * (y1 - y3))
        a = ((y2 - y3) * (x_arr - x3) + (x3 - x2) * (y_arr - y3)) / denominator
        b = ((y3 - y1) * (x_arr - x3) + (x1 - x3) * (y_arr - y3)) / denominator
        c = 1.0 - a - b

        #tf = (0.0 <= a <= 1.0 and 0.0 <= b <= 1.0 and 0.0 <= c <= 1.0)
        contains = np.logical_and(
            np.logical_and(0.0 <= a, a <= 1.0),
            np.logical_and(np.logical_and(0.0 <= b, b <= 1.0),
                           np.logical_and(0.0 <= c, c <= 1.0)))
        return contains

    def draw(self, viewer):
        cpoints = self.get_cpoints(viewer)
        cr = viewer.renderer.setup_cr(self)
        cr.draw_polygon(cpoints)

        if self.showcap:
            self.draw_caps(cr, self.cap, cpoints)


class XRange(Rectangle):
    """Draws an xrange on a DrawingCanvas.
    Parameters are:
    x1: start X coordinate in the data space
    x2: end X coordinate in the data space
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='x1', type=float, default=0.0, argpos=0,
                  description="First X coordinate of object"),
            Param(name='x2', type=float, default=0.0, argpos=1,
                  description="Second X coordinate of object"),
            Param(name='linewidth', type=int, default=0,
                  min=0, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='fillcolor', default='aquamarine',
                  valid=colors_plus_none, type=_color,
                  description="Color of fill"),
            Param(name='fillalpha', type=float, default=0.5,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of fill"),
            Param(name='drawdims', type=_bool,
                  default=False, valid=[False, True],
                  description="Annotate with dimensions of object"),
            Param(name='font', type=str, default='Sans Serif',
                  description="Font family for text"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        return cls(cxt.start_x, cxt.x, **cxt.drawparams)

    def __init__(self, x1, x2, color='yellow',
                 linewidth=0, linestyle='solid', showcap=False,
                 fillcolor='aquamarine', alpha=1.0,
                 drawdims=False, font='Sans Serif', fillalpha=0.5,
                 **kwdargs):
        Rectangle.__init__(self, x1, 0, x2, 0, color=color,
                           linewidth=linewidth,
                           linestyle=linestyle,
                           fill=True, fillcolor=fillcolor,
                           alpha=alpha, fillalpha=fillalpha,
                           drawdims=drawdims, font=font,
                           **kwdargs)
        self.kind = 'xrange'

    def contains_pts(self, pts):
        x_arr, y_arr = np.asarray(pts).T
        x1, y1, x2, y2 = self.get_llur()

        contains = np.logical_and(x1 <= x_arr, x_arr <= x2)
        return contains

    def get_edit_points(self, viewer):
        win_wd, win_ht = viewer.get_window_size()
        crdmap = viewer.get_coordmap('window')
        dx, dy = crdmap.to_data((0, win_ht // 2))
        pt = self.get_data_points(points=self.get_points())
        return [MovePoint((pt[0][0] + pt[2][0]) / 2.0, dy),
                EditPoint(pt[0][0], dy),
                EditPoint(pt[2][0], dy)]

    def draw(self, viewer):
        cr = viewer.renderer.setup_cr(self)

        win_wd, win_ht = viewer.get_window_size()
        cp = self.get_cpoints(viewer)
        cpoints = ((cp[0][0], 0), (cp[1][0], 0), (cp[2][0], win_ht), (cp[3][0], win_ht))

        cr.draw_polygon(cpoints)

        if self.drawdims:
            fontsize = self.scale_font(viewer)
            cr.set_font(self.font, fontsize, color=self.color)

            cx1, cy1 = cpoints[0]
            cx2, cy2 = cpoints[2]

            # draw label on X dimension
            cx = cx1 + (cx2 - cx1) // 2
            cy = (cy1 + cy2) // 2
            cr.draw_text(cx, cy, "%f:%f" % (self.x1, self.x2))


class YRange(Rectangle):
    """Draws a yrange on a DrawingCanvas.
    Parameters are:
    y1: start Y coordinate in the data space
    y2: end Y coordinate in the data space
    Optional parameters for linesize, color, etc.
    """

    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='coord', type=str, default='data',
                  valid=coord_names,
                  description="Set type of coordinates"),
            Param(name='y1', type=float, default=0.0, argpos=0,
                  description="First Y coordinate of object"),
            Param(name='y2', type=float, default=0.0, argpos=1,
                  description="Second Y coordinate of object"),
            Param(name='linewidth', type=int, default=0,
                  min=0, max=20, widget='spinbutton', incr=1,
                  description="Width of outline"),
            Param(name='linestyle', type=str, default='solid',
                  valid=['solid', 'dash'],
                  description="Style of outline (default solid)"),
            Param(name='color',
                  valid=colors_plus_none, type=_color, default='yellow',
                  description="Color of outline"),
            Param(name='alpha', type=float, default=1.0,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of outline"),
            Param(name='fillcolor', default='aquamarine',
                  valid=colors_plus_none, type=_color,
                  description="Color of fill"),
            Param(name='fillalpha', type=float, default=0.5,
                  min=0.0, max=1.0, widget='spinfloat', incr=0.05,
                  description="Opacity of fill"),
            Param(name='drawdims', type=_bool,
                  default=False, valid=[False, True],
                  description="Annotate with dimensions of object"),
            Param(name='font', type=str, default='Sans Serif',
                  description="Font family for text"),
        ]

    @classmethod
    def idraw(cls, canvas, cxt):
        return cls(cxt.start_y, cxt.y, **cxt.drawparams)

    def __init__(self, y1, y2, color='yellow',
                 linewidth=0, linestyle='solid', showcap=False,
                 fill=True, fillcolor='aquamarine', alpha=1.0,
                 drawdims=False, font='Sans Serif', fillalpha=0.5,
                 **kwdargs):
        Rectangle.__init__(self, 0, y1, 0, y2,
                           color=color, linewidth=linewidth,
                           linestyle=linestyle,
                           fill=True, fillcolor=fillcolor,
                           alpha=alpha, fillalpha=fillalpha,
                           drawdims=drawdims, font=font,
                           **kwdargs)
        self.kind = 'yrange'

    def contains_pts(self, pts):
        x_arr, y_arr = np.asarray(pts).T
        x1, y1, x2, y2 = self.get_llur()

        contains = np.logical_and(y1 <= y_arr, y_arr <= y2)
        return contains

    def get_edit_points(self, viewer):
        win_wd, win_ht = viewer.get_window_size()
        crdmap = viewer.get_coordmap('window')
        dx, dy = crdmap.to_data((win_wd // 2, 0))
        pt = self.get_data_points(points=self.get_points())
        return [MovePoint(dx, (pt[0][1] + pt[2][1]) / 2.0),
                EditPoint(dx, pt[0][1]),
                EditPoint(dx, pt[2][1])]

    def draw(self, viewer):
        cr = viewer.renderer.setup_cr(self)

        win_wd, win_ht = viewer.get_window_size()
        cp = self.get_cpoints(viewer)
        cpoints = ((0, cp[0][1]), (win_wd, cp[1][1]), (win_wd, cp[2][1]), (0, cp[3][1]))

        cr.draw_polygon(cpoints)

        if self.drawdims:
            fontsize = self.scale_font(viewer)
            cr.set_font(self.font, fontsize, color=self.color)

            cx1, cy1 = cpoints[0]
            cx2, cy2 = cpoints[2]

            # draw label on Y dimension
            cy = cy1 + (cy2 - cy1) // 2
            cx = (cx1 + cx2) // 2
            cr.draw_text(cx, cy, "%f:%f" % (self.y1, self.y2))


register_canvas_types(
    dict(text=Text, rectangle=Rectangle, circle=Circle,
         line=Line, point=Point, polygon=Polygon,
         freepolygon=FreePolygon, path=Path, freepath=FreePath,
         righttriangle=RightTriangle, triangle=Triangle,
         ellipse=Ellipse, square=Square, beziercurve=BezierCurve,
         box=Box, squarebox=SquareBox, xrange=XRange, yrange=YRange))

#END



"""Unit Tests for the Callbacks class."""

import pytest

import ginga.misc.Callback as Callback


class TestCallbacks(object):

    def test_init(self):
        test_callbacks = Callback.Callbacks()

        assert isinstance(test_callbacks.cb, dict)

        expected = 0
        actual = len(test_callbacks.cb)
        assert expected == actual

        expected = {}
        actual = test_callbacks.cb
        assert expected == actual

    def test_clear_callback_empties_list(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function():
            pass

        test_callbacks.cb["test_name"] = [(test_callbacks, (), {}), ]

        expected = 1
        actual = len(test_callbacks.cb["test_name"])
        assert expected == actual

        test_callbacks.clear_callback("test_name")

        expected = 0
        actual = len(test_callbacks.cb["test_name"])
        assert expected == actual

    def test_clear_callback_nonexistent_name(self):
        test_callbacks = Callback.Callbacks()

        assert "unknown_callback_key" not in test_callbacks.cb

        test_callbacks.clear_callback("unknown_callback_key")

        assert "unknown_callback_key" in test_callbacks.cb
        assert isinstance(test_callbacks.cb["unknown_callback_key"], list)

        expected = 0
        actual = len(test_callbacks.cb["unknown_callback_key"])
        assert expected == actual

    def test_enable_callback_nonexistent_name(self):
        test_callbacks = Callback.Callbacks()

        assert "unknown_callback_key" not in test_callbacks.cb

        test_callbacks.enable_callback("unknown_callback_key")

        assert "unknown_callback_key" in test_callbacks.cb
        assert isinstance(test_callbacks.cb["unknown_callback_key"], list)

        expected = 0
        actual = len(test_callbacks.cb["unknown_callback_key"])
        assert expected == actual

    def test_enable_callback_already_existent_name(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function():
            pass

        test_callbacks.cb["test_name"] = [(test_callbacks, (), {}), ]

        assert "test_name" in test_callbacks.cb
        assert isinstance(test_callbacks.cb["test_name"], list)

        expected = 1
        actual = len(test_callbacks.cb["test_name"])
        assert expected == actual

        test_callbacks.enable_callback("test_name")

        # testing that enable_callback() causes no change
        assert "test_name" in test_callbacks.cb
        assert isinstance(test_callbacks.cb["test_name"], list)

        expected = 1
        actual = len(test_callbacks.cb["test_name"])
        assert expected == actual

    def test_has_callback_existent_name(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function():
            pass

        test_callbacks.cb["test_name"] = [(test_callbacks, (), {}), ]

        expected = True
        actual = test_callbacks.has_callback("test_name")
        assert expected == actual

    def test_has_callback_non_existent_name(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function():
            pass

        test_callbacks.cb["test_name"] = [(test_callbacks, (), {}), ]

        expected = False
        actual = test_callbacks.has_callback("non_existent_name")
        assert expected == actual

    def test_has_callback_non_existent_name_empty_dict(self):
        test_callbacks = Callback.Callbacks()

        expected = False
        actual = test_callbacks.has_callback("non_existent_name")
        assert expected == actual

    def test_delete_callback_existent_name(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function():
            pass

        test_callbacks.cb["test_name"] = [(test_callbacks, (), {}), ]
        assert "test_name" in test_callbacks.cb

        test_callbacks.delete_callback("test_name")
        assert "test_name" not in test_callbacks.cb

    def test_delete_callback_non_existent_name(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function():
            pass

        test_callbacks.cb["test_name"] = [(test_callbacks, (), {}), ]

        with pytest.raises(Callback.CallbackError):
            test_callbacks.delete_callback("non_existent_name")

    def test_delete_callback_non_existent_name_empty_dict(self):
        test_callbacks = Callback.Callbacks()

        with pytest.raises(Callback.CallbackError):
            test_callbacks.delete_callback("non_existent_name")

    def test_add_callback(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function(obj, *args, **kwargs):
            pass

        test_callbacks.enable_callback("test_name")
        assert "test_name" in test_callbacks.cb

        test_callbacks.add_callback("test_name", test_callback_function)

        expected = 1
        actual = len(test_callbacks.cb["test_name"])
        assert expected == actual

        expected = (test_callback_function, (), {})
        actual = test_callbacks.cb["test_name"][0]
        assert expected == actual

        def another_test_callback_function(obj, *args, **kwargs):
            pass

        test_callbacks.add_callback(
            "test_name", another_test_callback_function)

        expected = 2
        actual = len(test_callbacks.cb["test_name"])
        assert expected == actual

        expected = (another_test_callback_function, (), {})
        actual = test_callbacks.cb["test_name"][1]
        assert expected == actual

    def test_add_callback_arguments(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function(obj, *args, **kwargs):
            pass

        test_callbacks.enable_callback("test_name")
        assert "test_name" in test_callbacks.cb

        test_callbacks.add_callback(
            "test_name",
            test_callback_function,
            'test_arg_1',
            'test_arg_2',
            test_keyword_arg1="test",
            test_keyword_arg2="test"
        )

        assert "test_name" in test_callbacks.cb

        expected = 1
        actual = len(test_callbacks.cb["test_name"])
        assert expected == actual

        expected = (
            test_callback_function,
            ('test_arg_1', 'test_arg_2'),
            {'test_keyword_arg1': 'test', 'test_keyword_arg2': 'test'}
        )
        actual = test_callbacks.cb["test_name"][0]
        assert expected == actual

    def test_add_callback_exception(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function(obj, *args, **kwargs):
            pass

        with pytest.raises(Callback.CallbackError):
            test_callbacks.add_callback("test_name", test_callback_function)

    def test_set_callback(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function(obj, *args, **kwargs):
            pass

        test_callbacks.set_callback("test_name", test_callback_function)

        assert "test_name" in test_callbacks.cb

        expected = 1
        actual = len(test_callbacks.cb["test_name"])
        assert expected == actual

        expected = (test_callback_function, (), {})
        actual = test_callbacks.cb["test_name"][0]
        assert expected == actual

        def another_test_callback_function(obj, *args, **kwargs):
            pass

        test_callbacks.set_callback(
            "test_name", another_test_callback_function)

        expected = 2
        actual = len(test_callbacks.cb["test_name"])
        assert expected == actual

        expected = (another_test_callback_function, (), {})
        actual = test_callbacks.cb["test_name"][1]
        assert expected == actual

    def test_set_callback_arguments(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function(obj, *args, **kwargs):
            pass

        test_callbacks.set_callback(
            "test_name",
            test_callback_function,
            'test_arg_1',
            'test_arg_2',
            test_keyword_arg1="test",
            test_keyword_arg2="test"
        )

        assert "test_name" in test_callbacks.cb

        expected = 1
        actual = len(test_callbacks.cb["test_name"])
        assert expected == actual

        expected = (
            test_callback_function,
            ('test_arg_1', 'test_arg_2'),
            {'test_keyword_arg1': 'test', 'test_keyword_arg2': 'test'}
        )
        actual = test_callbacks.cb["test_name"][0]
        assert expected == actual

    def test_make_callback_non_existent_name(self):
        test_callbacks = Callback.Callbacks()

        expected = None
        actual = test_callbacks.make_callback("non_existent_event_name")

        assert expected == actual

    def test_make_callback_empty_callback_list(self):
        test_callbacks = Callback.Callbacks()

        test_callbacks.enable_callback("known_name")

        assert "known_name" in test_callbacks.cb

        expected = False
        actual = test_callbacks.make_callback("known_name")

        assert expected == actual

    def test_make_callback_single_callback_true(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function(obj, *args, **kwargs):
            return True

        test_callbacks.set_callback("test_name", test_callback_function)

        expected = True
        actual = test_callbacks.make_callback("test_name")
        assert expected == actual

    def test_make_callback_single_callback_false(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function(obj, *args, **kwargs):
            return False

        test_callbacks.set_callback("test_name", test_callback_function)

        expected = False
        actual = test_callbacks.make_callback("test_name")
        assert expected == actual

    def test_make_callback_multiple_callback_all_true(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function(obj, *args, **kwargs):
            return True

        def another_test_callback_function(obj, *args, **kwargs):
            return True

        test_callbacks.set_callback("test_name", test_callback_function)
        test_callbacks.set_callback(
            "test_name", another_test_callback_function)

        expected = True
        actual = test_callbacks.make_callback("test_name")
        assert expected == actual

    def test_make_callback_multiple_callback_some_true(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function(obj, *args, **kwargs):
            return False

        def another_test_callback_function(obj, *args, **kwargs):
            return True

        test_callbacks.set_callback("test_name", test_callback_function)
        test_callbacks.set_callback(
            "test_name", another_test_callback_function)

        expected = True
        actual = test_callbacks.make_callback("test_name")
        assert expected == actual

    def test_make_callback_multiple_callback_all_false(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function(obj, *args, **kwargs):
            return False

        def another_test_callback_function(obj, *args, **kwargs):
            return False

        test_callbacks.set_callback("test_name", test_callback_function)
        test_callbacks.set_callback(
            "test_name", another_test_callback_function)

        expected = False
        actual = test_callbacks.make_callback("test_name")
        assert expected == actual

    def test_make_callback_raises_no_exception(self):
        test_callbacks = Callback.Callbacks()

        # This function when used as a callback should raise a TypeError
        # as the callbacks, from the logic in ginga.misc.Callback.Callbacks
        # always take the calling object as the first argument
        def test_callback_function():
            return True

        test_callbacks.set_callback("test_name", test_callback_function)

        # Checking that the callback eats up the TypeError exception
        expected = False
        actual = test_callbacks.make_callback("test_name")
        assert expected == actual

    def test_make_callback_raises_no_exception_completes_all_callbacks(self):
        test_callbacks = Callback.Callbacks()

        def test_callback_function():
            return True

        def another_test_callback_function(obj, *args, **kwargs):
            return True

        test_callbacks.set_callback("test_name", test_callback_function)
        test_callbacks.set_callback(
            "test_name", another_test_callback_function)

        # Checking that the callback eats up the TypeError exception and
        # continues to the other callback and returns True in the end
        expected = True
        actual = test_callbacks.make_callback("test_name")
        assert expected == actual

# END



#
# CanvasObject.py -- base class for shapes drawn on ginga canvases.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np
from collections import namedtuple

from ginga.misc import Callback, Bunch
from ginga import trcalc, colors

from . import coordmap

__all__ = ['CanvasObjectBase', 'get_canvas_type', 'get_canvas_types',
           'register_canvas_type', 'register_canvas_types']

colors_plus_none = [None] + colors.get_colors()

coord_names = ['data', 'wcs', 'cartesian', 'percentage', 'window']

Point = namedtuple('Point', ['x', 'y'])


class EditPoint(Point):
    edit_color = 'yellow'


class MovePoint(EditPoint):
    edit_color = 'orangered'


class ScalePoint(EditPoint):
    edit_color = 'green'


class RotatePoint(EditPoint):
    edit_color = 'skyblue'


class CanvasObjectError(Exception):
    pass


class CanvasObjectBase(Callback.Callbacks):
    """This is the abstract base class for a CanvasObject.  A CanvasObject
    is an item that can be placed on a ImageViewCanvas.

    This class defines common methods used by all such objects.
    """

    def __init__(self, **kwdargs):
        if not hasattr(self, 'cb'):
            Callback.Callbacks.__init__(self)
        self.cap = 'ball'
        self.cap_radius = 4
        self.editable = True
        self.pickable = False
        self.coord = 'data'
        self.ref_obj = None
        self.__dict__.update(kwdargs)
        self.data = None
        self.crdmap = None
        self.tag = None
        if not hasattr(self, 'kind'):
            self.kind = None
        # For debugging
        self.name = None
        self.viewer = None

        # For callbacks
        for name in ('edited', 'pick-down', 'pick-move', 'pick-up',
                     'pick-hover', 'pick-enter', 'pick-leave',
                     'pick-key'):
            self.enable_callback(name)

    def initialize(self, canvas, viewer, logger):
        self.viewer = viewer
        self.logger = logger
        if self.crdmap is None:
            if self.coord is None:
                # default mapping is to data coordinates
                self.coord = 'data'

            if self.coord == 'offset':
                self.crdmap = coordmap.OffsetMapper(viewer, self.ref_obj)
            else:
                self.crdmap = viewer.get_coordmap(self.coord)

    def sync_state(self):
        """This method called when changes are made to the parameters.
        subclasses should override if they need any special state handling.
        """
        pass

    def set_data(self, **kwdargs):
        if self.data is None:
            self.data = Bunch.Bunch(kwdargs)
        else:
            self.data.update(kwdargs)

    def get_data(self, *args):
        if len(args) == 0:
            return self.data
        elif len(args) == 1:
            return self.data[args[0]]
        elif len(args) == 2:
            try:
                return self.data[args[0]]
            except KeyError:
                return args[1]
        else:
            raise CanvasObjectError(
                "method get_data() takes at most 2 arguments")

    def use_coordmap(self, mapobj):
        self.crdmap = mapobj

    def canvascoords(self, viewer, data_x, data_y, center=None):
        if center is not None:
            self.logger.warn(
                "`center` keyword is ignored and will be deprecated")

        return viewer.get_canvas_xy(data_x, data_y)

    def is_compound(self):
        return False

    def contains_pts(self, points):
        contains = np.asarray([False] * len(points))
        return contains

    def contains_pt(self, pt):
        pts = np.asarray([pt])
        return self.contains_pts(pts)[0]

    def select_contains_pt(self, viewer, pt):
        return self.contains_pt(pt)

    def draw_arrowhead(self, cr, x1, y1, x2, y2):
        i1, j1, i2, j2 = self.calc_vertexes(x1, y1, x2, y2)

        alpha = getattr(self, 'alpha', 1.0)
        cr.set_fill(self.color, alpha=alpha)
        cr.draw_polygon(((x2, y2), (i1, j1), (i2, j2)))
        cr.set_fill(None)

    def draw_caps(self, cr, cap, points, radius=None):
        i = 0
        for pt in points:
            cx, cy = pt[:2]
            if radius is None:
                radius = self.cap_radius
            alpha = getattr(self, 'alpha', 1.0)
            if cap == 'ball':
                color = self.color
                # Draw edit control points in different colors than the others
                if isinstance(pt, EditPoint):
                    cr.set_fill('black', alpha=alpha)
                    cr.draw_circle(cx, cy, radius + 2.0)

                    color = pt.edit_color

                cr.set_fill(color, alpha=alpha)
                cr.draw_circle(cx, cy, radius)
                #cr.set_fill(self, None)
            i += 1

    def draw_edit(self, cr, viewer):
        bbox = self.get_bbox()
        cpoints = self.get_cpoints(viewer, points=bbox, no_rotate=True)
        cr.set_fill(None)
        cr.set_line(color='cyan', style='dash')
        cr.draw_polygon(cpoints)

        points = self.get_edit_points(viewer)
        cpoints = self.get_cpoints(viewer, points=points)

        # preserve point types for coloring
        def _map_cpt(pt, cpt):
            if isinstance(pt, EditPoint):
                return pt.__class__(*cpt)
            return cpt

        cpoints = tuple([_map_cpt(points[i], cpoints[i])
                         for i in range(len(points))])
        self.draw_caps(cr, 'ball', cpoints)

    def calc_radius(self, viewer, p1, p2):
        x1, y1 = p1[:2]
        x2, y2 = p2[:2]
        # TODO: the accuracy of this calculation of radius might be improved?
        radius = np.sqrt(abs(y2 - y1)**2 + abs(x2 - x1)**2)
        return (x1, y1, radius)

    def calc_vertexes(self, start_cx, start_cy, end_cx, end_cy,
                      arrow_length=10, arrow_degrees=0.35):

        angle = np.arctan2(end_cy - start_cy, end_cx - start_cx) + np.pi

        cx1 = end_cx + arrow_length * np.cos(angle - arrow_degrees)
        cy1 = end_cy + arrow_length * np.sin(angle - arrow_degrees)
        cx2 = end_cx + arrow_length * np.cos(angle + arrow_degrees)
        cy2 = end_cy + arrow_length * np.sin(angle + arrow_degrees)

        return (cx1, cy1, cx2, cy2)

    def swapxy(self, x1, y1, x2, y2):
        if x2 < x1:
            x1, x2 = x2, x1
        if y2 < y1:
            y1, y2 = y2, y1
        return (x1, y1, x2, y2)

    def scale_font(self, viewer):
        scale = viewer.get_scale_max()
        basesize = getattr(self, 'fontsize', 10.0)
        if basesize is None:
            basesize = 10.0
        min_size = getattr(self, 'fontsize_min', 2.0)
        n = 1.4
        fontsize = max(min_size, basesize + np.log(scale) / np.log(n))
        max_size = getattr(self, 'fontsize_max', None)
        if max_size is not None:
            fontsize = min(max_size, fontsize)
        return fontsize

    def get_points(self):
        """Get the set of points that is used to draw the object.

        Points are returned in *data* coordinates.
        """
        if hasattr(self, 'points'):
            points = self.crdmap.to_data(self.points)
        else:
            points = []
        return points

    def get_data_points(self, points=None):
        """Points returned are in data coordinates."""
        if points is None:
            points = self.points
        points = self.crdmap.to_data(points)
        return points

    def set_data_points(self, points):
        """
        Input `points` must be in data coordinates, will be converted
        to the coordinate space of the object and stored.
        """
        self.points = np.asarray(self.crdmap.data_to(points))

    def rotate_deg(self, thetas, offset):
        points = np.asarray(self.get_data_points(), dtype=np.double)
        points = trcalc.rotate_coord(points, thetas, offset)
        self.set_data_points(points)

    def rotate_by_deg(self, thetas):
        ref_pt = self.get_reference_pt()
        self.rotate_deg(thetas, ref_pt)

    def rerotate_by_deg(self, thetas, detail):
        ref_pt = detail.center_pos
        points = np.asarray(detail.points, dtype=np.double)
        points = trcalc.rotate_coord(points, thetas, ref_pt)
        self.set_data_points(points)

    def move_delta_pt(self, off_pt):
        points = np.asarray(self.get_data_points(), dtype=np.double)
        points = np.add(points, off_pt)
        self.set_data_points(points)

    def move_to_pt(self, dst_pt):
        ref_pt = self.get_reference_pt()
        off_pt = np.subtract(dst_pt, ref_pt)
        self.move_delta_pt(off_pt)

    def get_num_points(self):
        return(len(self.points))

    def set_point_by_index(self, i, pt):
        #self.points[i] = self.crdmap.data_to(pt)
        # Can we eventually use something like the above?
        points = np.asarray(self.points, dtype=np.float)
        points[i] = self.crdmap.data_to(pt)
        self.points = points

    def get_point_by_index(self, i):
        return self.crdmap.to_data(self.points[i])

    def scale_by_factors(self, factors):
        ctr_pt = self.get_center_pt()
        pts = np.asarray(self.get_data_points(), dtype=np.double)
        pts = np.add(np.multiply(np.subtract(pts, ctr_pt), factors), ctr_pt)
        self.set_data_points(pts)

    def rescale_by_factors(self, factors, detail):
        ctr_pt = detail.center_pos
        pts = np.asarray(detail.points, dtype=np.double)
        pts = np.add(np.multiply(np.subtract(pts, ctr_pt), factors), ctr_pt)
        self.set_data_points(pts)

    def setup_edit(self, detail):
        """subclass should override as necessary."""
        detail.center_pos = self.get_center_pt()

    def calc_rotation_from_pt(self, pt, detail):
        x, y = pt[:2]
        ctr_x, ctr_y = detail.center_pos[:2]
        start_x, start_y = detail.start_pos[:2]
        # calc angle of starting point wrt origin
        deg1 = np.degrees(np.arctan2(start_y - ctr_y,
                                     start_x - ctr_x))
        # calc angle of current point wrt origin
        deg2 = np.degrees(np.arctan2(y - ctr_y, x - ctr_x))
        delta_deg = deg2 - deg1
        return delta_deg

    def calc_scale_from_pt(self, pt, detail):
        x, y = pt[:2]
        ctr_x, ctr_y = detail.center_pos[:2]
        start_x, start_y = detail.start_pos[:2]
        dx, dy = start_x - ctr_x, start_y - ctr_y
        # calc distance of starting point wrt origin
        dist1 = np.sqrt(dx**2.0 + dy**2.0)
        dx, dy = x - ctr_x, y - ctr_y
        # calc distance of current point wrt origin
        dist2 = np.sqrt(dx**2.0 + dy**2.0)
        scale_f = dist2 / dist1
        return scale_f

    def calc_dual_scale_from_pt(self, pt, detail):
        x, y = pt[:2]
        ctr_x, ctr_y = detail.center_pos[:2]
        start_x, start_y = detail.start_pos[:2]
        # calc distance of starting point wrt origin
        dx, dy = start_x - ctr_x, start_y - ctr_y
        # calc distance of current point wrt origin
        ex, ey = x - ctr_x, y - ctr_y
        scale_x, scale_y = float(ex) / dx, float(ey) / dy
        return scale_x, scale_y

    def convert_mapper(self, tomap):
        """
        Converts our object from using one coordinate map to another.

        NOTE: In some cases this only approximately preserves the
        equivalent point values when transforming between coordinate
        spaces.
        """
        frommap = self.crdmap
        if frommap == tomap:
            return

        # mild hack to convert radii on objects that have them
        if hasattr(self, 'radius'):
            # get coordinates of a point radius away from center
            # under current coordmap
            x0, y0 = frommap.offset_pt((self.x, self.y), (self.radius, 0))
            pts = frommap.to_data(((self.x, self.y), (x0, y0)))
            pts = tomap.data_to(pts)
            self.radius = np.fabs(pts[1][0] - pts[0][0])

        elif hasattr(self, 'xradius'):
            # similar to above case, but there are 2 radii
            x0, y0 = frommap.offset_pt((self.x, self.y), (self.xradius,
                                                          self.yradius))
            pts = frommap.to_data(((self.x, self.y), (x0, y0)))
            pts = tomap.data_to(pts)
            self.xradius = np.fabs(pts[1][0] - pts[0][0])
            self.yradius = np.fabs(pts[1][1] - pts[0][1])

        data_pts = self.get_data_points()

        # set our map to the new map
        self.crdmap = tomap

        self.set_data_points(data_pts)

    def point_within_radius(self, points, pt, canvas_radius,
                            scales=(1.0, 1.0)):
        """Points `points` and point `pt` are in data coordinates.
        Return True for points within the circle defined by
        a center at point `pt` and within canvas_radius.
        """
        scale_x, scale_y = scales
        x, y = pt
        a_arr, b_arr = np.asarray(points).T
        dx = np.fabs(x - a_arr) * scale_x
        dy = np.fabs(y - b_arr) * scale_y
        new_radius = np.sqrt(dx**2 + dy**2)
        res = (new_radius <= canvas_radius)
        return res

    def within_radius(self, viewer, points, pt, canvas_radius):
        """Points `points` and point `pt` are in data coordinates.
        Return True for points within the circle defined by
        a center at point `pt` and within canvas_radius.
        The distance between points is scaled by the canvas scale.
        """
        scales = viewer.get_scale_xy()
        return self.point_within_radius(points, pt, canvas_radius,
                                        scales)

    def get_pt(self, viewer, points, pt, canvas_radius=None):
        """Takes an array of points `points` and a target point `pt`.
        Returns the first index of the point that is within the
        radius of the target point.  If none of the points are within
        the radius, returns None.
        """
        if canvas_radius is None:
            canvas_radius = self.cap_radius

        if hasattr(self, 'rot_deg'):
            # rotate point back to cartesian alignment for test
            ctr_pt = self.get_center_pt()
            pt = trcalc.rotate_coord(pt, [-self.rot_deg], ctr_pt)

        res = self.within_radius(viewer, points, pt, canvas_radius)
        return np.flatnonzero(res)

    def point_within_line(self, points, p_start, p_stop, canvas_radius):
        # TODO: is there an algorithm with the cross and dot products
        # that is more efficient?
        r = canvas_radius
        x1, y1 = p_start
        x2, y2 = p_stop
        xmin, xmax = min(x1, x2) - r, max(x1, x2) + r
        ymin, ymax = min(y1, y2) - r, max(y1, y2) + r
        div = np.sqrt((x2 - x1)**2 + (y2 - y1)**2)

        a_arr, b_arr = np.asarray(points).T
        d = np.fabs((x2 - x1) * (y1 - b_arr) - (x1 - a_arr) * (y2 - y1)) / div

        contains = np.logical_and(
            np.logical_and(xmin <= a_arr, a_arr <= xmax),
            np.logical_and(d <= canvas_radius,
                           np.logical_and(ymin <= b_arr, b_arr <= ymax)))
        return contains

    def within_line(self, viewer, points, p_start, p_stop, canvas_radius):
        """Points `points` and line endpoints `p_start`, `p_stop` are in
        data coordinates.
        Return True for points within the line defined by a line from
        p_start to p_end and within `canvas_radius`.
        The distance between points is scaled by the viewer's canvas scale.
        """
        scale_x, scale_y = viewer.get_scale_xy()
        new_radius = canvas_radius * 1.0 / min(scale_x, scale_y)

        return self.point_within_line(points, p_start, p_stop, new_radius)

    def get_center_pt(self):
        """Return the geometric average of points as data_points.
        """
        P = np.asarray(self.get_data_points(), dtype=np.double)
        x = P[:, 0]
        y = P[:, 1]
        ctr_x = np.sum(x) / float(len(x))
        ctr_y = np.sum(y) / float(len(y))
        return ctr_x, ctr_y

    def get_reference_pt(self):
        return self.get_center_pt()

    def get_move_scale_rotate_pts(self, viewer):
        """Returns 3 edit control points for editing this object: a move
        point, a scale point and a rotate point.  These points are all in
        data coordinates.
        """
        scale = viewer.get_scale_min()
        ref_x, ref_y = self.get_center_pt()
        xl, yl, xu, yu = self.get_llur()
        offset = 8.0 / scale
        scl_x, scl_y = xl - offset, yl - offset
        rot_x, rot_y = xu + offset, yu + offset
        if hasattr(self, 'rot_deg'):
            # if this is an object with a rotation attribute, pre rotate
            # the control points in the opposite direction, because they
            # will be rotated back
            theta = -self.rot_deg
            scl_x, scl_y = trcalc.rotate_pt(scl_x, scl_y, theta,
                                            xoff=ref_x, yoff=ref_y)
            rot_x, rot_y = trcalc.rotate_pt(rot_x, rot_y, theta,
                                            xoff=ref_x, yoff=ref_y)
        move_pt = MovePoint(ref_x, ref_y)
        scale_pt = ScalePoint(scl_x, scl_y)
        rotate_pt = RotatePoint(rot_x, rot_y)

        return (move_pt, scale_pt, rotate_pt)

    def get_cpoints(self, viewer, points=None, no_rotate=False):
        # If points are passed, they are assumed to be in data space
        if points is None:
            points = self.get_points()

        if (not no_rotate) and hasattr(self, 'rot_deg') and self.rot_deg != 0.0:
            # rotate vertices according to rotation
            ctr_x, ctr_y = self.get_center_pt()
            points = trcalc.rotate_coord(points, [self.rot_deg], (ctr_x, ctr_y))

        crdmap = viewer.get_coordmap('native')
        return crdmap.data_to(points)

    def get_bbox(self, points=None):
        """
        Get bounding box of this object.

        Returns
        -------
        (p1, p2, p3, p4): a 4-tuple of the points in data coordinates,
        beginning with the lower-left and proceeding counter-clockwise.
        """
        if points is None:
            x1, y1, x2, y2 = self.get_llur()
            return ((x1, y1), (x1, y2), (x2, y2), (x2, y1))
        else:
            return trcalc.strip_z(trcalc.get_bounds(points))

    # --- TO BE DEPRECATED METHODS ---

    def move_delta(self, xoff, yoff):
        """For backward compatibility.  TO BE DEPRECATED--DO NOT USE.
        Use move_delta_pt instead.
        """
        self.move_delta_pt((xoff, yoff))

    def move_to(self, xdst, ydst):
        """For backward compatibility.  TO BE DEPRECATED--DO NOT USE.
        Use move_to_pt() instead.
        """
        self.move_to_pt((xdst, ydst))

    def scale_by(self, scale_x, scale_y):
        """For backward compatibility.  TO BE DEPRECATED--DO NOT USE.
        Use scale_by_factors() instead.
        """
        self.scale_by_factors((scale_x, scale_y))

    def rescale_by(self, scale_x, scale_y, detail):
        """For backward compatibility.  TO BE DEPRECATED--DO NOT USE.
        Use rescale_by_factors() instead.
        """
        self.rescale_by_factors((scale_x, scale_y), detail)

    def rotate(self, theta_deg, xoff=0, yoff=0):
        """For backward compatibility.  TO BE DEPRECATED--DO NOT USE.
        Use rotate_deg() instead.
        """
        self.rotate_deg([theta_deg], (xoff, yoff))

    def rotate_by(self, theta_deg):
        """For backward compatibility.  TO BE DEPRECATED--DO NOT USE.
        Use rotate_by_deg() instead.
        """
        self.rotate_by_deg([theta_deg])

    def contains_arr(self, x_arr, y_arr):
        """For backward compatibility.  TO BE DEPRECATED--DO NOT USE.
        Use contains_pts() instead.
        """
        pts = np.asarray((x_arr, y_arr)).T
        return self.contains_pts(pts)

    def contains(self, x, y):
        """For backward compatibility.  TO BE DEPRECATED--DO NOT USE.
        Use contains_pt() instead.
        """
        return self.contains_pt((x, y))

    def select_contains(self, viewer, x, y):
        """For backward compatibility.  TO BE DEPRECATED--DO NOT USE.
        Use select_contains_pt() instead.
        """
        return self.select_contains_pt(viewer, (x, y))


# this is the data structure to which drawing classes are registered
drawCatalog = Bunch.Bunch(caseless=True)


def get_canvas_types():
    # force registration of all canvas types
    import ginga.canvas.types.all  # noqa

    return drawCatalog


def get_canvas_type(name):
    # force registration of all canvas types
    import ginga.canvas.types.all  # noqa

    return drawCatalog[name]


def register_canvas_type(name, klass):
    global drawCatalog
    drawCatalog[name] = klass


def register_canvas_types(klass_dict):
    global drawCatalog
    drawCatalog.update(klass_dict)


# funky boolean converter
_bool = lambda st: str(st).lower() == 'true'  # noqa

# color converter
_color = lambda name: name  # noqa

# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
The ``Colorbar`` plugin shows a colorbar indicating the colormap applied
to the image and showing the example values along the range.

**Plugin Type: Global**

``Colorbar`` is a global plugin.  Only one instance can be opened.

**Usage**

Clicking and dragging in the ``Colorbar`` window will shift the colormap
left or right.  Scrolling will stretch or shrink the colormap at the
cursor position.  Right-clicking will restore the colormap from any
shift or stretch.

If the focus shifts to another channel, the colorbar will be updated
to reflect that channel's colormap and value information.

"""
from ginga import GingaPlugin
from ginga.misc import Bunch
from ginga.gw import ColorBar

__all__ = ['Colorbar']


class Colorbar(GingaPlugin.GlobalPlugin):

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super(Colorbar, self).__init__(fv)

        self._image = None
        self.active = None
        self.info = None

        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Colorbar')
        self.settings.add_defaults(cbar_height=36, fontsize=10)
        self.settings.load(onError='silent')

        fv.add_callback('add-channel', self.add_channel_cb)
        fv.add_callback('delete-channel', self.delete_channel_cb)
        fv.add_callback('channel-change', self.change_cbar)
        self.colorbar = None
        self.cursor_obj = None
        self.gui_up = False

    def build_gui(self, container):
        cbar = ColorBar.ColorBar(self.logger, settings=self.settings)
        cbar.set_cmap(self.fv.cm)
        cbar.set_imap(self.fv.im)
        cbar_w = cbar.get_widget()
        cbar_ht = self.settings.get('cbar_height', 36)
        cbar_w.resize(-1, cbar_ht)

        self.colorbar = cbar
        cbar.add_callback('motion', self.cbar_value_cb)

        # see cbar_val_cb()
        if self.fv.gpmon.has_plugin('Cursor'):
            self.cursor_obj = self.fv.gpmon.get_plugin('Cursor')

        container.add_widget(cbar_w, stretch=0)
        self.gui_up = True

    def add_channel_cb(self, viewer, channel):
        settings = channel.settings
        settings.get_setting('cuts').add_callback(
            'set', self.change_range_cb, channel.fitsimage)

        chname = channel.name
        info = Bunch.Bunch(chname=chname, channel=channel)
        channel.extdata._colorbar_info = info

        fi = channel.fitsimage
        rgbmap = fi.get_rgbmap()
        rgbmap.add_callback('changed', self.rgbmap_cb, channel)

    def delete_channel_cb(self, viewer, channel):
        chname = channel.name
        self.logger.debug("deleting channel %s" % (chname))
        self.active = None
        self.info = None

    def _match_cmap(self, fitsimage, colorbar):
        """
        Help method to change the ColorBar to match the cut levels or
        colormap used in a ginga ImageView.
        """
        rgbmap = fitsimage.get_rgbmap()
        loval, hival = fitsimage.get_cut_levels()
        colorbar.set_range(loval, hival)
        # If we are sharing a ColorBar for all channels, then store
        # to change the ColorBar's rgbmap to match our
        colorbar.set_rgbmap(rgbmap)

    def change_cbar(self, viewer, channel):
        if self.gui_up and channel is not None:
            self._match_cmap(channel.fitsimage, self.colorbar)

    def change_range_cb(self, setting, value, fitsimage):
        """
        This method is called when the cut level values (lo/hi) have
        changed in a channel.  We adjust them in the ColorBar to match.
        """
        if not self.gui_up:
            return
        if fitsimage != self.fv.getfocus_viewer():
            # values have changed in a channel that doesn't have the focus
            return False
        loval, hival = value
        self.colorbar.set_range(loval, hival)

    def cbar_value_cb(self, cbar, value, event):
        """
        This method is called when the user moves the mouse over the
        ColorBar.  It displays the value of the mouse position in the
        ColorBar in the Readout (if any).
        """
        if self.cursor_obj is not None:
            readout = self.cursor_obj.readout
            if readout is not None:
                maxv = readout.maxv
                text = "Value: %-*.*s" % (maxv, maxv, value)
                readout.set_text(text)

    def rgbmap_cb(self, rgbmap, channel):
        """
        This method is called when the RGBMap is changed.  We update
        the ColorBar to match.
        """
        if not self.gui_up:
            return
        fitsimage = channel.fitsimage
        if fitsimage != self.fv.getfocus_fitsimage():
            return False
        self.change_cbar(self.fv, channel)

    def start(self):
        channel = self.fv.get_channel_info()
        self.change_cbar(self.fv, channel)

    def stop(self):
        self.gui_up = False
        self.cursor_obj = None
        self.colorbar = None
        return True

    def close(self):
        self.fv.stop_global_plugin(str(self))
        return True

    def __str__(self):
        return 'colorbar'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_Colorbar', package='ginga')

# END



#
# ImageViewQtGL.py -- a backend for Ginga using Qt's QGLWidget
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from io import BytesIO

from ginga.qtw.QtHelp import QtCore
from ginga.qtw import ImageViewQt
from ginga import Mixins, Bindings
from ginga.canvas import transform

# Local imports
from .CanvasRenderGL import CanvasRenderer

# GL imports
# TODO: find how to import this from qtpy
from PyQt5.QtOpenGL import QGLWidget as QOpenGLWidget


class ImageViewQtGLError(ImageViewQt.ImageViewQtError):
    pass


class RenderGLWidget(QOpenGLWidget):

    def __init__(self, *args, **kwdargs):
        QOpenGLWidget.__init__(self, *args, **kwdargs)

        self.viewer = None

    def initializeGL(self):
        self.viewer.renderer.gl_initialize()

    def resizeGL(self, width, height):
        self.viewer.configure_window(width, height)

    def paintGL(self):
        self.viewer.renderer.gl_paint()

    def sizeHint(self):
        width, height = 300, 300
        if self.viewer is not None:
            width, height = self.viewer.get_desired_size()
        return QtCore.QSize(width, height)


class ImageViewQtGL(ImageViewQt.ImageViewQt):

    def __init__(self, logger=None, rgbmap=None, settings=None):
        ImageViewQt.ImageViewQt.__init__(self, logger=logger,
                                         rgbmap=rgbmap, settings=settings)

        self.imgwin = RenderGLWidget()
        self.imgwin.viewer = self
        # Qt expects 32bit BGRA data for color images
        self.rgb_order = 'RGBA'

        self.renderer = CanvasRenderer(self)

        # we replace two transforms in the catalog for OpenGL rendering
        #self.trcat['WindowNativeTransform'] = WindowGLTransform
        self.trcat['WindowNativeTransform'] = \
            transform.CartesianWindowTransform.inverted_class()
        self.trcat['CartesianNativeTransform'] = transform.PassThruTransform
        self.recalc_transforms()

    def configure_window(self, width, height):
        self.logger.debug("window size reconfigured to %dx%d" % (
            width, height))
        self.renderer.resize((width, height))

        self.configure(width, height)

    def get_rgb_image_as_widget(self):
        return self.imgwin.grabFrameBuffer()

    def get_rgb_image_as_buffer(self, output=None, format='png',
                                quality=90):
        ibuf = output
        if ibuf is None:
            ibuf = BytesIO()

        qimg = self.get_rgb_image_as_widget()
        qimg.save(ibuf, format=format, quality=quality)
        return ibuf

    def update_image(self):
        if self.imgwin is None:
            return

        self.logger.debug("updating window")
        self.imgwin.update()

    def gl_update(self):
        self.imgwin.update()


class RenderGLWidgetZoom(ImageViewQt.RenderMixin, RenderGLWidget):
    pass


class ImageViewEvent(ImageViewQtGL, ImageViewQt.QtEventMixin):

    def __init__(self, logger=None, rgbmap=None, settings=None):
        ImageViewQtGL.__init__(self, logger=logger, rgbmap=rgbmap,
                               settings=settings)

        # replace the widget our parent provided
        imgwin = RenderGLWidgetZoom()

        imgwin.viewer = self
        self.imgwin = imgwin

        ImageViewQt.QtEventMixin.__init__(self)


class ImageViewZoom(Mixins.UIMixin, ImageViewEvent):

    # class variables for binding map and bindings can be set
    bindmapClass = Bindings.BindingMapper
    bindingsClass = Bindings.ImageViewBindings

    @classmethod
    def set_bindingsClass(cls, klass):
        cls.bindingsClass = klass

    @classmethod
    def set_bindmapClass(cls, klass):
        cls.bindmapClass = klass

    def __init__(self, logger=None, settings=None, rgbmap=None,
                 bindmap=None, bindings=None):
        ImageViewEvent.__init__(self, logger=logger, settings=settings,
                                rgbmap=rgbmap)
        Mixins.UIMixin.__init__(self)

        self.ui_set_active(True)

        if bindmap is None:
            bindmap = ImageViewZoom.bindmapClass(self.logger)
        self.bindmap = bindmap
        bindmap.register_for_events(self)

        if bindings is None:
            bindings = ImageViewZoom.bindingsClass(self.logger)
        self.set_bindings(bindings)

    def get_bindmap(self):
        return self.bindmap

    def get_bindings(self):
        return self.bindings

    def set_bindings(self, bindings):
        self.bindings = bindings
        bindings.set_bindings(self)


class CanvasView(ImageViewZoom):

    def __init__(self, logger=None, settings=None, rgbmap=None,
                 bindmap=None, bindings=None):
        ImageViewZoom.__init__(self, logger=logger, settings=settings,
                               rgbmap=rgbmap,
                               bindmap=bindmap, bindings=bindings)

        # Needed for UIMixin to propagate events correctly
        self.objects = [self.private_canvas]

    def set_canvas(self, canvas, private_canvas=None):
        super(CanvasView, self).set_canvas(canvas,
                                           private_canvas=private_canvas)

        self.objects[0] = self.private_canvas


#END
